import json

import click

import globus_sdk.exc
from globus_sdk import TransferClient
from globus_sdk.authorizers import ClientCredentialsAuthorizer

from gcs_cli.config import (
    GCS_CLIENT_DATA_CONFIG_KEY,
    internal_auth_client,
    token_storage_adapter,
)
from gcs_cli.helpers import do_link_auth_flow
from gcs_cli.parsing import EndpointIDorLocalhost, globus_command

INFO_FILE = "/var/lib/globus-connect-server/info.json"

_LOGGED_IN_RESPONSE = """\
You are already logged into a GCS endpoint.

You may logout with
    globus-connect-server logout

Or force a new login with
  globus-connect-server login --force ENDPOINT_ID
"""


_INFO_FILE_NOT_FOUND_ERR = """\
The 'info.json' file was not found.

This suggests that you are attempting to log in on a system where
Globus Connect Server is not configured.

You must either install and configure Globus Connect Server to perform
a 'localhost' login, or supply the fully-qualified DNS name of the
Globus Connect Server endpoint you wish to use."""

_INFO_FILE_PERMISSION_DENIED_ERR = """\
Permission Denied while attempting to read {}

This likely means that something has gone wrong with your Globus
Connect Server configuration.
You may want to explicitly change permissions on the info.json file
to make it world readable.

Alternatively, you may be able to complete a login flow by supplying
the fully-qualified DNS name of the Globus Connect Server endpoint
you wish to use.""".format(
    INFO_FILE
)


def _collect_login_info(ep_id_or_localhost):
    if ep_id_or_localhost == "localhost":
        try:
            with open(INFO_FILE, "r") as fp:
                epdata = json.load(fp)
        except FileNotFoundError:
            raise click.ClickException(_INFO_FILE_NOT_FOUND_ERR)
        except OSError as err:
            if err.errno == 13:
                raise click.ClickException(_INFO_FILE_PERMISSION_DENIED_ERR)
            raise
        return (epdata["endpoint_id"], epdata["client_id"], epdata["domain_name"])
    else:
        # get client_id and fqdn from Transfer
        internal_authorizer = ClientCredentialsAuthorizer(
            internal_auth_client(), "urn:globus:auth:scope:transfer.api.globus.org:all"
        )
        transfer_client = TransferClient(authorizer=internal_authorizer)
        try:
            endpoint_info = transfer_client.get_endpoint(ep_id_or_localhost)
        except globus_sdk.exc.TransferAPIError:
            raise click.ClickException(
                "Either no such Endpoint exists, or it is private\n\n"
                "To login to a private Endpoint, you must run\n"
                "  globus-connect-server login localhost\n"
                "from one of the Endpoint hosts"
            )

        gcs_client_id = endpoint_info["owner_id"]
        gcs_fqdn = endpoint_info["DATA"][0]["hostname"]

        return (ep_id_or_localhost, gcs_client_id, gcs_fqdn)


@globus_command(
    "login",
    short_help="Get credentials for a GCS endpoint.",
    disable_options=["format", "explicit_host"],
)
@click.option(
    "--force",
    is_flag=True,
    help="Do a fresh auth flow, ignoring any existing credentials",
)
@click.argument("endpoint_id", type=EndpointIDorLocalhost())
def login_command(endpoint_id, force):
    """
    Do an auth flow to get credentials for a GCS endpoint. To revoke the credentials log
    out of the CLI with `globus-connect-server logout`.
    """
    endpoint_id, gcs_client_id, gcs_fqdn = _collect_login_info(endpoint_id)

    # if not forcing, stop if user already logged in
    if not force and check_logged_in():
        click.echo(_LOGGED_IN_RESPONSE.format(endpoint_id))
        return

    # use a link login
    do_link_auth_flow(gcs_client_id)

    adapter = token_storage_adapter()
    adapter.store_config(
        GCS_CLIENT_DATA_CONFIG_KEY, {"client_id": gcs_client_id, "fqdn": gcs_fqdn}
    )

    # print epilog
    click.echo(
        "You have successfully logged into GCS endpoint "
        "{} at {}!".format(gcs_client_id, gcs_fqdn)
    )


def check_logged_in():
    """
    perform a series of checks to get a complete/working "logged in" state
    if any check fails, return False
    """
    adapter = token_storage_adapter()
    # first, get the GCS client info (so we can lookup tokens)
    gcs_client_data = adapter.read_config(GCS_CLIENT_DATA_CONFIG_KEY)
    if not gcs_client_data:
        return False

    # try to get the tokens from config
    gcs_client_id = gcs_client_data["client_id"]
    token_data = adapter.read_as_dict().get(gcs_client_id)
    if not token_data:
        return False

    # pull out the refresh token from there (at this point, the storage adapter
    # should be giving us a complete/working response unless someone messed
    # with the DB, but let's play it safe and check for null
    refresh_token = token_data.get("refresh_token")
    if refresh_token is None:
        return False

    # get or create the instance client
    auth_client = internal_auth_client()

    # check that token and client are valid
    try:
        res = auth_client.oauth2_validate_token(refresh_token)
        if not res["active"]:
            return False

    # if the instance client is invalid, an AuthAPIError will be raised
    # we then force a new client to be created before continuing
    except globus_sdk.exc.AuthAPIError:
        return False

    return True
