import glob
import json
import os
import shlex
import shutil
import socket
import subprocess
import textwrap

import click
import requests

from gcs_cli.helpers import is_logged_in
from gcs_cli.parsing import globus_command
from gcs_cli.sdk_client.host_header_ssl_adapter import HostHeaderSSLAdapter


def _run_diagnostic_step(cmd, force_try_all):
    stagename = (
        str(cmd)
        if not callable(cmd)
        else getattr(cmd, "stagename", "func:{}()".format(cmd.__name__))
    )
    click.secho("== diagnostic: {} ==".format(stagename), fg="yellow", bold=True)

    try:
        if callable(cmd):
            result = cmd()
            if isinstance(result, str):
                click.echo(result)
            return

        if isinstance(cmd, str):
            cmd = shlex.split(cmd)
        arg0 = cmd[0]
        if force_try_all or shutil.which(arg0):
            try:
                subprocess.run(cmd, timeout=30)
            except subprocess.TimeoutExpired:
                click.secho("command timed out", fg="red", bold=True)
        else:
            click.secho("{} was not found in the PATH".format(arg0), fg="red")
    # at the very least, subprocess execution can raise an OSError
    # when attempting to execute a non-executable file or when the command name
    # can't be resolved
    # any errors here should be caught to avoid crashing the whole diagnostic run
    except Exception as e:
        click.secho("command failed: {}".format(e), fg="red", bold=True)


def _is_debian():
    """return true if the platform is a debian-like system
    false if it is not or if detection fails in any way"""
    try:
        with open("/etc/os-release", "r") as f:
            for line in f:
                if line.strip() == "ID=debian" or line.strip() == "ID_LIKE=debian":
                    return True
    except Exception:
        pass
    return False


def cat(path, wildcard=False):
    def show_step():
        if not wildcard:
            files = [path]
        else:
            files = glob.glob(path)

        for filename in files:
            click.echo("cat " + filename)
            if os.path.exists(filename):
                with open(filename) as f:
                    click.echo(textwrap.indent(f.read(), "  "))
            else:
                click.echo("no file named {}".format(filename))

    show_step.stagename = "func:cat({})".format(path)

    return show_step


def systemctl_status(service):
    return "systemctl status {} --no-pager --lines 10".format(service)


def get_fqdn():
    return socket.getfqdn()


def user_is_logged_in():
    return str(is_logged_in())


def running_as_root():
    return os.geteuid() == 0


def get_service_names(force_try_all):
    services = [
        "gcs_manager",
        "gcs_manager.socket",
        "gcs_manager_assistant",
        "globus-gridftp-server",
    ]
    # add apache2 or httpd depending on platform in use
    if force_try_all:
        return ["apache2", "httpd"] + services
    if _is_debian():
        return ["apache2"] + services
    return ["httpd"] + services


def get_info_uri():
    try:
        with open("/var/lib/globus-connect-server/info.json", "r") as f:
            info = json.load(f)
            return "https://{}/api/info".format(info["domain_name"])
    except Exception as e:
        return click.style(
            "Parsing info.json failed: {}".format(e), fg="red", bold=True
        )


def resolve_nodes():
    try:
        with open("/var/lib/globus-connect-server/info.json", "r") as f:
            info = json.load(f)
    except Exception as e:
        return click.style(
            "Parsing info.json failed: {}".format(e), fg="red", bold=True
        )
    domain_name = info["domain_name"]

    res = []
    for addr in socket.getaddrinfo(domain_name, 443, proto=socket.IPPROTO_TCP):
        res.append(addr[4][0])

    if len(res) == 0:
        res.append(
            click.style(
                "Domain {} does not resolve to any addresses".format(domain_name),
                fg="red",
                bold=True,
            )
        )
    return "Domain {} resolves to {}".format(domain_name, ", ".join(res))


def contact_info_uri():
    try:
        with open("/var/lib/globus-connect-server/info.json", "r") as f:
            info = json.load(f)
    except Exception as e:
        return click.style(
            "Parsing info.json failed: {}".format(e), fg="red", bold=True
        )

    domain_name = info["domain_name"]
    client = requests.sessions.Session()
    client.mount("https://", HostHeaderSSLAdapter())

    res = []
    for addr in socket.getaddrinfo(domain_name, 443, proto=socket.IPPROTO_TCP):
        ip_address = addr[4][0]
        try:
            info_data = client.get(
                "https://{}/api/info".format(ip_address),
                headers={"Host": domain_name},
                timeout=1,
            )
            info_data.raise_for_status()
            for datum in info_data.json()["data"]:
                if datum["DATA_TYPE"].startswith("info#"):
                    res.append(
                        click.style(
                            "{}: {}".format(ip_address, json.dumps(datum, indent=2))
                        )
                    )
        except Exception as e:
            res.append(
                "Error accessing GCS Manager at {}: {}".format(
                    ip_address, click.style(str(e), fg="red", bold=True)
                )
            ),
    return "\n".join(res)


@globus_command(
    "self-diagnostic", disable_options=["map_http_status", "format", "explicit_host"]
)
@click.option(
    "--force-try-all",
    is_flag=True,
    help="Don't skip any steps that are expected to fail",
)
def self_diagnostic_command(force_try_all):
    """Run several diagnostic utilities which can be used to provide feedback to the
    Globus Team.

    This tool may produce a large amount of output.
    """
    any_user_diagnostics = [
        "uname -a",
        "sestatus",
        "ifconfig",
        get_fqdn,
        cat("/etc/os-release"),
        cat("/var/lib/globus-connect-server/info.json"),
        user_is_logged_in,
        "globus-gridftp-server --version",
        get_info_uri,
        resolve_nodes,
        contact_info_uri,
    ] + [systemctl_status(x) for x in get_service_names(force_try_all)]
    root_user_diagnostics = [
        cat("/etc/gridftp.conf"),
        cat("/etc/gridftp.d/*", wildcard=True),
    ]

    for cmd in any_user_diagnostics:
        _run_diagnostic_step(cmd, force_try_all)

    is_root = running_as_root()
    click.secho("= check running_as_root:{} =".format(is_root), fg="yellow", bold=True)
    if is_root or force_try_all:
        for cmd in root_user_diagnostics:
            _run_diagnostic_step(cmd, force_try_all)
