import uuid

import click

from gcs_cli.config import GCS_CLIENT_DATA_CONFIG_KEY, token_storage_adapter
from gcs_cli.helpers import do_link_auth_flow, requires_login
from gcs_cli.parsing import globus_command
from gcs_cli.services.auth import get_client


@globus_command(
    "update",
    short_help="Update your CLI auth session",
    disable_options=["format", "map_http_status", "explicit_host"],
)
@click.argument("identities", nargs=-1, required=False)
@click.option(
    "--all", is_flag=True, help="authenticate with every identity in your identity set"
)
@click.option("--mfa", is_flag=True, help="require multi-factor authentication")
@requires_login
def session_update(identities, all, mfa):
    """
    Update your current CLI auth session by authenticating
    with specific identities.
    """
    adapter = token_storage_adapter()
    gcs_client_data = adapter.read_config(GCS_CLIENT_DATA_CONFIG_KEY)
    gcs_client_id = gcs_client_data["client_id"]

    if (not (identities or all)) or (identities and all):
        raise click.UsageError("Either give one or more IDENTITIES or use --all")
    auth_client = get_client()

    # if --all use every identity id in the user's identity set
    if all:
        res = auth_client.oauth2_userinfo()
        try:
            identity_ids = [user["sub"] for user in res["identity_set"]]
        except KeyError:
            click.echo(
                "Your current login does not have the consents required "
                "to view your full identity set. Please log in again "
                "to agree to the required consents.",
                err=True,
            )
            click.get_current_context().exit(1)

    # otherwise try to resolve any non uuid values to identity ids
    else:
        identity_ids = []
        identity_names = []

        for val in identities:
            try:
                uuid.UUID(val)
                identity_ids.append(val)
            except ValueError:
                identity_names.append(val)

        if identity_names:
            res = auth_client.get_identities(usernames=identity_names)["identities"]

            for name in identity_names:
                for identity in res:
                    if identity["username"] == name:
                        identity_ids.append(identity["id"])
                        break
                else:
                    click.echo("No such identity {}".format(val), err=True)
                    click.get_current_context().exit(1)

    # create session params once we have all identity ids
    session_params = {
        "session_required_identities": ",".join(identity_ids),
        "session_message": "Authenticate to update your CLI session.",
    }

    if mfa:
        session_params["session_required_mfa"] = True

    # use a link login if remote session or user requested
    do_link_auth_flow(gcs_client_id, session_params=session_params)

    click.echo(
        "\nYou have successfully updated your CLI session.\n"
        "Use 'globus-connect-server session show' to see the updated session."
    )
