import click

from gcs_cli.helpers import requires_login
from gcs_cli.parsing import collection_id_arg, globus_command
from gcs_cli.prettyprint import FORMAT_TEXT_RECORD, formatted_print
from gcs_cli.sdk_client.data import SharingPolicyDocument
from gcs_cli.services.gcs import get_client


@globus_command("create", short_help="Create a new Sharing Policy")
@collection_id_arg
@click.option(
    "--user",
    "-u",
    metavar="USER",
    multiple=True,
    required=True,
    help=(
        "Restrict this policy to the given user. This may be passed multiple "
        "times to apply the policy to multiple users"
    ),
)
@click.option(
    "--read",
    metavar="PATH",
    multiple=True,
    help=(
        "Restrict sharing to be read only for the given path. This may be "
        "passed multiple times to allow multiple read path restrictions"
    ),
)
@click.option(
    "--read-write",
    metavar="PATH",
    multiple=True,
    help=(
        "Restrict sharing to be read/write for the given path. This may be "
        "passed multiple times to allow multiple read/write path restrictions"
    ),
)
@click.option(
    "--none",
    metavar="PATH",
    multiple=True,
    help=(
        "Restrict sharing to be disallowed for the given path. This may be "
        "passed multiple times to allow multiple restrictions"
    ),
)
@requires_login
def sharing_policy_create(
    collection_id=None,
    user=None,
    read=None,
    read_write=None,
    none=None,
):
    """
    Create a new Sharing Policy. Each policy applies to a single mapped
    collection, and can restrict or allow paths to be shared by a set of users.
    Sharing path restrictions may include read, read_write, or no access. At
    least one path restriction (of any type) must be passed on the
    command-line.

    To change the default sharing policies for all users, update the
    sharing_restrict_paths property of the collection using the
    globus-connect-server collection update command.
    """

    if not read and not read_write and not none:
        click.echo("Missing path restrictions")
        click.get_current_context().exit(1)

    sharing_policy_doc = SharingPolicyDocument(
        collection_id=str(collection_id),
        users=user,
        read=list(read),
        read_write=list(read_write),
        none=list(none),
    )

    client = get_client()
    res = client.create_sharing_policy(sharing_policy_doc)

    # output
    formatted_print(
        res, fields=[("Sharing Policy ID", "id")], text_format=FORMAT_TEXT_RECORD
    )
