import logging.config
import os
import stat

import click
from globus_sdk_tokenstorage import SQLiteAdapter

import globus_sdk

__all__ = [
    # option name constants
    "GLOBUS_ENV",
    "GLOBUS_PROFILE",
    "internal_native_client",
    "internal_auth_client",
    "token_storage_adapter",
]

# configuration names stored and retrieved on the globus_sdk_tokenstorage.SQLiteAdapter
# using store_config and read_config
AUTH_CLIENT_DATA_CONFIG_KEY = "auth_client_data"
# TODO: right now, this uses a singular "gcs_client_data" because the CLI only
# talks to one GCS at a time
# this is baked into login and logout, but it doesn't really belong there in
# this way
# instead, we need to be storing and using multiple resource servers' config data (for
# now, just FQDN) -- so perhaps this will become a family of string and we'll
# store under
#   "gcs_client_data_{client_id}"
# as the key?
#
# the impact on logout, in particular, needs to be figured out
GCS_CLIENT_DATA_CONFIG_KEY = "gcs_client_data"

# get the environment from env var (not exported)
GLOBUS_ENV = os.environ.get("GLOBUS_SDK_ENVIRONMENT")

# get profile name (allows you to sign in with multiple users)
GLOBUS_PROFILE = os.environ.get("GLOBUS_PROFILE", "default")


# config for use whenever we need to do things using
# instance clients from the CLI Native App Template
# primarily accessed via `internal_auth_client()`
def _get_template_client_id():
    default_id = "d736c0c7-264a-4bb3-afb6-1cf6cab99f90"
    return {
        "sandbox": "0d551879-cc99-491b-b80f-edc33493c208",
        "integration": "462ba401-cfe9-4ce7-a628-0e0173bd6a7c",
        "test": "facf8603-76fc-4e42-907c-566c9f4f022a",
        "staging": "7bacbcf4-e7d1-4794-9aca-b432f6c4d491",
        "preview": "d72b7d65-1385-4559-880d-e0cc2be2b0be",
    }.get(GLOBUS_ENV, default_id)


def _get_tokenstorage_db():
    storagedb_dir = os.path.expanduser("~/.globus/gcs_cli/")
    os.makedirs(storagedb_dir, exist_ok=True, mode=stat.S_IRWXU)
    return os.path.join(storagedb_dir, "tokenstore.db")


def _get_tokenstorage_namespace():
    if "/" in GLOBUS_PROFILE:
        raise ValueError("Not allowed to have '/' in GLOBUS_PROFILE!")
    if GLOBUS_ENV:
        return GLOBUS_ENV + "/" + GLOBUS_PROFILE
    else:
        return GLOBUS_PROFILE


def _check_token_permissions(dbname):
    if click.get_current_context().command.name == "logout":
        return

    bad_permissions = False

    dirname = os.path.dirname(dbname)

    if os.path.exists(dirname):
        dir_stat = os.stat(dirname)
        if dir_stat.st_mode & (stat.S_IRWXO | stat.S_IRWXG):
            bad_permissions = True
    if os.path.exists(dbname):
        db_stat = os.stat(dbname)
        if db_stat.st_mode & (stat.S_IRWXO | stat.S_IRWXG):
            bad_permissions = True

    if bad_permissions:
        raise click.ClickException(
            "Your token storage database has insecure file permissions.\n"
            "Do the following to invalidate tokens and fix the permissions:\n\n"
            "globus-connect-server logout\n"
            "chmod 700 ~/.globus ~/.globus/gcs_cli\n"
            "rm ~/.globus/gcs_cli/tokenstore.db\n"
        )


def token_storage_adapter():
    if not hasattr(token_storage_adapter, "_instance"):
        # namespace is equal to the current environment
        tokendb = token_storage_adapter._instance = SQLiteAdapter(
            _get_tokenstorage_db(), namespace=_get_tokenstorage_namespace()
        )
        _check_token_permissions(tokendb.dbname)
    return token_storage_adapter._instance


def load_client_data():
    adapter = token_storage_adapter()
    conf = adapter.read_config(AUTH_CLIENT_DATA_CONFIG_KEY)
    return conf or {}


def internal_native_client():
    template_id = load_client_data().get("template_id", _get_template_client_id())
    return globus_sdk.NativeAppAuthClient(template_id)


def internal_auth_client(force_new_client=False):
    """
    Looks up the values for this CLI's Instance Client in config

    If none exists or force_new_client is True,
    registers a new Instance Client with GLobus Auth

    Returns a ConfidentialAppAuthClient
    """
    client_data = load_client_data()
    client_id = client_data.get("client_id")
    client_secret = client_data.get("client_secret")
    template_id = client_data.get("template_id", _get_template_client_id())
    template_client = internal_native_client()
    existing = client_id and client_secret

    # if we are forcing a new client, delete any existing client
    if force_new_client and existing:
        existing_client = globus_sdk.ConfidentialAppAuthClient(client_id, client_secret)
        try:
            existing_client.delete("/v2/api/clients/{}".format(client_id))

        # if the client secret has been invalidated or the client has
        # already been removed, we continue on
        except globus_sdk.exc.AuthAPIError:
            pass

    # if we require a new client to be made
    if force_new_client or not existing:
        # register a new instance client with auth
        body = {"client": {"template_id": template_id, "name": "GCS CLI"}}
        res = template_client.post("/v2/api/clients", json_body=body)

        # get values and write to config
        credential_data = res["included"]["client_credential"]
        client_id = client_data["client_id"] = credential_data["client"]
        client_secret = client_data["client_secret"] = credential_data["secret"]
        token_storage_adapter().store_config(AUTH_CLIENT_DATA_CONFIG_KEY, client_data)

    return globus_sdk.ConfidentialAppAuthClient(
        client_id, client_secret, app_name="Globus Connect Server CLI"
    )


def setup_logging(level="DEBUG"):
    conf = {
        "version": 1,
        "formatters": {
            "basic": {"format": "[%(levelname)s] %(name)s::%(funcName)s() %(message)s"}
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "level": level,
                "formatter": "basic",
            }
        },
        "loggers": {"globus_sdk": {"level": level, "handlers": ["console"]}},
    }

    logging.config.dictConfig(conf)
