import re

from OpenSSL import crypto


def load_certificate(path):
    """
    Load the X.509 certificate from path, returning a normalized output
    containing just the PEM-encoded data

    :param path:
        Path to the certificate file.
    """
    with open(path, "r") as f:
        data = f.read()
    cert = crypto.load_certificate(crypto.FILETYPE_PEM, data)
    return crypto.dump_certificate(crypto.FILETYPE_PEM, cert).decode("ascii")


def load_certificate_chain(path):
    """
    Load a list of X.509 certificates from path, returning a normalized output
    containing just the PEM-encoded data

    :param path:
        Path to the certificate chain file.
    """
    if path:
        certs = []
        start = r"-----BEGIN CERTIFICATE-----[\r]?\n"
        end = r"-----END CERTIFICATE-----[\r]?[\n]?"
        with open(path, "r") as f:
            data = f.read()
        while data:
            startp = re.search(start, data, re.MULTILINE)
            endp = re.search(end, data, re.MULTILINE)
            if startp and endp and startp.start() < endp.end():
                cert = crypto.load_certificate(
                    crypto.FILETYPE_PEM, data[startp.start() : endp.end()]
                )
                certs.append(
                    crypto.dump_certificate(crypto.FILETYPE_PEM, cert).decode("ascii")
                )
                data = data[endp.end() :]
            else:
                break
        return "\n".join(certs)


def load_private_key(path, passphrase=None):
    """
    Load a private key from path, returning a normalized output
    containing just the PEM-encoded data, using the optional passphrase to
    decrypt the data if needed.

    :param path:
        Path to the private key file.
    :param passphrase:
        String or callable to return a passphrase
    """
    with open(path, "r") as f:
        data = f.read()

    private_key = crypto.load_privatekey(
        crypto.FILETYPE_PEM, data, passphrase=passphrase
    )
    return crypto.dump_privatekey(crypto.FILETYPE_PEM, private_key).decode("ascii")
