"""
Data helper classes for constructing GCS API documents. All classes should
extend ``dict``, so they can be passed seamlessly to
:class:`GCSClient` methods without conversion.
"""
from __future__ import unicode_literals

from globus_sdk.base import safe_stringify


def _set_raw_kwargs(d, **kwargs):
    for arg, val in kwargs.items():
        if val is not None:
            d[arg] = val


def _set_bool_kwargs(d, **kwargs):
    for arg, val in kwargs.items():
        if val is not None:
            d[arg] = bool(val)


def _set_string_kwargs(d, **kwargs):
    for arg, val in kwargs.items():
        if val is not None:
            d[arg] = safe_stringify(val)


def _set_string_list_kwargs(d, **kwargs):
    for arg, val in kwargs.items():
        if val is not None:
            d[arg] = [safe_stringify(x) for x in val]


class StorageGatewayDocument(dict):
    """
    Convenience class for constructing a Storage Gateway document
    to use as the `data` parameter to :meth:`create_storage_gateway` or
    :meth:`update_storage_gateway`

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``display_name`` (*string*)
      Name of the Storage Gateway.

    ``connector_id`` (*string*)
      UUID of the connector type that this Storage Gateway interacts with.

    ``identity_mappings`` (*list of dicts*)
      A list of IdentityMapping objects which are applied to user identities to attempt
      to determine what accounts are available for access.

    ``policies`` (*dict*)
      Connector specific storage policies. It is recommended that
      you use one of the Policy helper classes (e.g. `POSIXStoragePolicies`
      if you are using the POSIX connector) to create these.

    ``allowed_domains`` (*list of string*)
      List of allowed domains. Users creating credentials or collections on
      this Storage Gateway must have an identity in one of these domains.

    ``restrict_paths`` (*dict*)
      Path restrictions within this Storage Gateway. [Private]

    ``high_assurance`` (*bool*)
      Flag indicating if the Storage Gateway requires high assurance features.

    ``authentication_timeout_mins`` (*int*)
      Timeout (in minutes) during which a user is required to have authenticated
      in a session to access this storage gateway.
    """

    def __init__(
        self,
        DATA_TYPE="storage_gateway#1.1.0",
        display_name=None,
        connector_id=None,
        root=None,
        identity_mappings=None,
        policies=None,
        allowed_domains=None,
        high_assurance=None,
        require_mfa=None,
        authentication_timeout_mins=None,
        users_allow=None,
        users_deny=None,
        **kwargs
    ):
        _set_string_kwargs(
            self,
            DATA_TYPE=DATA_TYPE,
            display_name=display_name,
            connector_id=connector_id,
            root=root,
        )
        _set_raw_kwargs(
            self,
            identity_mappings=identity_mappings,
            policies=policies,
            high_assurance=high_assurance,
            require_mfa=require_mfa,
            authentication_timeout_mins=authentication_timeout_mins,
            users_allow=users_allow,
            users_deny=users_deny,
        )
        _set_string_list_kwargs(self, allowed_domains=allowed_domains)

        self.update(kwargs)


class POSIXStoragePolicies(dict):
    """
    Convenience class for constructing a POSIX Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``groups_allow`` (*list of string*)
      List of POSIX group IDs allowed to access this Storage Gateway. [Private]

    ``groups_deny`` (*list of string*)
      List of POSIX group IDs denied access this Storage Gateway. [Private]
    """

    def __init__(
        self,
        DATA_TYPE="posix_storage_policies#1.0.0",
        groups_allow=None,
        groups_deny=None,
        **kwargs
    ):

        self["DATA_TYPE"] = safe_stringify(DATA_TYPE)

        _set_string_list_kwargs(
            self, groups_allow=groups_allow, groups_deny=groups_deny
        )

        self.update(kwargs)


class POSIXStagingStoragePolicies(POSIXStoragePolicies):
    """
    Convenience class for constructing a POSIX Staging Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``groups_allow`` (*list of string*)
      List of POSIX group IDs allowed to access this Storage Gateway. [Private]

    ``groups_deny`` (*list of string*)
      List of POSIX group IDs denied access this Storage Gateway. [Private]

    ``stage_app`` (*string*)
      Path to the stage app.[Private]

    ``environment`` (*list of string*)
      Variables to set in the environment when executing the stage_app.[Private]
    """

    def __init__(
        self,
        DATA_TYPE="posix_staging_storage_policies#1.0.0",
        stage_app=None,
        environment=None,
        **kwargs
    ):
        if environment is not None:
            self["environment"] = [
                {
                    "name": safe_stringify(env["name"]),
                    "value": safe_stringify(env["value"]),
                }
                for env in environment
            ]

        if stage_app is not None:
            _set_string_kwargs(self, stage_app=stage_app)

        super().__init__(DATA_TYPE=DATA_TYPE, **kwargs)


class BlackPearlStoragePolicies(dict):
    """
    Convenience class for constructing a BlackPearl Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``s3_endpoint`` (*string*)
      The URL of the S3 endpoint of the BlackPearl appliance to use to access
      collections on this Storage Gateway.

    ``bp_access_id_file`` (*string*)
      Path to the file which provides mappings from usernames within the
      configured identity domain to the ID and secret associated with the user's
      BlackPearl account
    """

    def __init__(
        self,
        DATA_TYPE="blackpearl_storage_policies#1.0.0",
        s3_endpoint=None,
        bp_access_id_file=None,
        **kwargs
    ):
        _set_string_kwargs(
            self,
            DATA_TYPE=DATA_TYPE,
            s3_endpoint=s3_endpoint,
            bp_access_id_file=bp_access_id_file,
        )

        self.update(kwargs)


class BoxStoragePolicies(dict):
    """
    Convenience class for constructing a Box Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``enterpriseID`` (*string*)
      Identifies which Box Enterprise this Storage Gateway is authorized access
      to. [Private]

    ``boxAppSettings`` (*dict*)
      Values that the Storage Gateway uses to identify and authenticate the
      with the Box API. [Private]
    """

    def __init__(
        self,
        DATA_TYPE="box_storage_policies#1.0.0",
        enterpriseID=None,
        boxAppSettings=None,
        **kwargs
    ):
        _set_string_kwargs(self, DATA_TYPE=DATA_TYPE, enterpriseID=enterpriseID)
        _set_raw_kwargs(self, boxAppSettings=boxAppSettings)

        self.update(kwargs)


class CephStoragePolicies(dict):
    """
    Convenience class for constructing a Ceph Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``s3_endpoint`` (*string*)
      URL of the S3 API endpoint

    ``s3_buckets`` (*list of string*)
      List of buckets not owned by the collection owner that will be shown in
      the root of collections created at the base of this Storage Gateway.

    ``ceph_admin_key_id`` (*string*)
      Administrator key id used to authenticate with the ceph admin service to
      obtain user credentials. [Private]

    ``ceph_admin_secret_key`` (*string*)
      Administrator secret key used to authenticate with the ceph admin service
      to obtain user credentials. [Private]
    """

    def __init__(
        self,
        DATA_TYPE="ceph_storage_policies#1.0.0",
        s3_endpoint=None,
        s3_buckets=None,
        ceph_admin_key_id=None,
        ceph_admin_secret_key=None,
        **kwargs
    ):
        _set_string_kwargs(
            self,
            DATA_TYPE=DATA_TYPE,
            s3_endpoint=s3_endpoint,
            ceph_admin_key_id=ceph_admin_key_id,
            ceph_admin_secret_key=ceph_admin_secret_key,
        )
        _set_string_list_kwargs(self, s3_buckets=s3_buckets)

        self.update(kwargs)


class GoogleDriveStoragePolicies(dict):
    """
    Convenience class for constructing a Google Drive Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``client_id`` (*string*)
      Client ID registered with the Google Application console to access Google
      Drive. [Private]

    ``secret`` (*string*)
      Secret created to access access Google Drive with the client_id in this
      policy. [Private]

    ``user_api_rate_quota`` (*integer*)
      User API Rate quota associated with this client ID.
    """

    def __init__(
        self,
        DATA_TYPE="google_drive_storage_policies#1.0.0",
        client_id=None,
        secret=None,
        user_api_rate_quota=None,
        **kwargs
    ):
        _set_string_kwargs(
            self, DATA_TYPE=DATA_TYPE, client_id=client_id, secret=secret
        )
        _set_raw_kwargs(self, user_api_rate_quota=user_api_rate_quota)

        self.update(kwargs)


class GoogleCloudStoragePolicies(dict):
    """
    Convenience class for constructing a Google Cloud Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``client_id`` (*string*)
      Client ID registered with the Google Application console to access Google
      Drive. [Private]

    ``secret`` (*string*)
      Secret created to access access Google Drive with the client_id in this
      policy. [Private]

    ``service_account_key`` (*dict*)
      Credentials for use with service account auth, read from a
      Google-provided json file. [Private]

    ``buckets`` (*list of string*)
      The list of Google Cloud Storage buckets which the Storage Gateway is
      allowed to access, as well as the list of buckets that will be shown in
      root level directory listings. If this list is unset, bucket access is
      unrestricted and all non public credential accessible buckets will be
      shown in root level directory listings. The value is a list of bucket
      names.

    ``projects`` (*list of string*)
      The list of Google Cloud Storage projects which the Storage Gateway is
      allowed to access. If this list is unset, project access is unrestricted.
      The value is a list of project id strings.
    """

    def __init__(
        self,
        DATA_TYPE="google_cloud_storage_policies#1.0.0",
        client_id=None,
        secret=None,
        service_account_key=None,
        buckets=None,
        projects=None,
        **kwargs
    ):
        _set_string_kwargs(
            self, DATA_TYPE=DATA_TYPE, client_id=client_id, secret=secret
        )
        _set_string_list_kwargs(self, buckets=buckets, projects=projects)
        _set_raw_kwargs(self, service_account_key=service_account_key)

        self.update(kwargs)


class OneDriveStoragePolicies(dict):
    """
    Convenience class for constructing a OneDrive Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``client_id`` (*string*)
      Client ID registered with the MS Application console to access
      OneDrive. [Private]

    ``secret`` (*string*)
      Secret created to access access MS with the client_id in this
      policy. [Private]

    ``tenant`` (*string*)
      MS Tenant ID from which to allow user logins. [Private]

    ``user_api_rate_limit`` (*integer*)
      User API Rate limit associated with this client ID.
    """

    def __init__(
        self,
        DATA_TYPE="onedrive_storage_policies#1.0.0",
        client_id=None,
        secret=None,
        tenant=None,
        user_api_rate_limit=None,
        **kwargs
    ):
        _set_string_kwargs(
            self, DATA_TYPE=DATA_TYPE, client_id=client_id, secret=secret, tenant=tenant
        )
        _set_raw_kwargs(self, user_api_rate_limit=user_api_rate_limit)

        self.update(kwargs)


class AzureBlobStoragePolicies(dict):
    """
    Convenience class for constructing a Azure Blob Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``client_id`` (*string*)
      Client ID registered with the MS Application console to access
      Azure Blob [Private]

    ``secret`` (*string*)
      Secret created to access access MS with the client_id in this
      policy. [Private]

    ``tenant`` (*string*)
      MS Tenant ID from which to allow user logins. [Private]

    ``account`` (*string*)
      Azure Storage account. [Private]

    ``auth_type`` (*string*)
      Auth type: user, service_principal or user_service_principal

    ``adls`` (*bool*)
      ADLS support enabled or not. [Private]
    """

    def __init__(
        self,
        DATA_TYPE="azure_blob_storage_policies#1.0.0",
        client_id=None,
        secret=None,
        tenant=None,
        account=None,
        auth_type=None,
        adls=None,
        **kwargs
    ):
        _set_string_kwargs(
            self,
            DATA_TYPE=DATA_TYPE,
            client_id=client_id,
            secret=secret,
            tenant=tenant,
            account=account,
            auth_type=auth_type,
        )
        _set_raw_kwargs(self, adls=adls)

        self.update(kwargs)


class S3StoragePolicies(dict):
    """
    Convenience class for constructing a Google Cloud Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``s3_endpoint`` (*string*)
      URL of the S3 API endpoint

    ``s3_buckets`` (*list of string*)
      List of buckets not owned by the collection owner that will be shown in
      the root of collections created at the base of this Storage Gateway.

    ``s3_user_credential_required`` (*bool*)
      Flag indicating if a Globus User must register a user credential in order
      to create a Guest Collection on this Storage Gateway.
    """

    def __init__(
        self,
        DATA_TYPE="s3_storage_policies#1.0.0",
        s3_endpoint=None,
        s3_buckets=None,
        s3_user_credential_required=None,
        **kwargs
    ):
        _set_string_kwargs(self, DATA_TYPE=DATA_TYPE, s3_endpoint=s3_endpoint)
        _set_raw_kwargs(self, s3_user_credential_required=s3_user_credential_required)
        _set_string_list_kwargs(self, s3_buckets=s3_buckets)

        self.update(kwargs)


class ActiveScaleStoragePolicies(S3StoragePolicies):
    """
    ActiveScale Storage Policy uses S3 policy DATA_TYPE
    """

    pass


class IrodsStoragePolicies(dict):
    """
    Convenience class for constructing an iRODS Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``irods_environment_file`` (*string*)
      Path to iRODS environent file on the endpoint

    ``irods_authentication_file`` (*string*)
      Path to iRODS authentication file on the endpoint
    """

    def __init__(
        self,
        DATA_TYPE="irods_storage_policies#1.0.0",
        irods_environment_file=None,
        irods_authentication_file=None,
        **kwargs
    ):
        _set_string_kwargs(
            self,
            DATA_TYPE=DATA_TYPE,
            irods_environment_file=irods_environment_file,
            irods_authentication_file=irods_authentication_file,
        )

        self.update(kwargs)


class CollectionDocument(dict):
    """
    Convenience class for constructing a Collection document
    to use as the `data` parameter to :meth:`update_collection`

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``storage_gateway_id`` (*string*)
      Unique ID of the Storage Gateway which this collection provides access to.
      This value can not change after the collection is created.

    ``identity_id`` (*string*)
      Globus Auth identity to who acts as the owner of this collection. This
      identity is an `administrator` on the collection.

    ``collection_type`` (*string*)
      Either "guest" or "mapped". A mapped collection requires an account on the
      system to access the administrator-defined collection. A guest collection
      allows users to share access to their data on a Storage Gateway by
      registering an credential with the GCS Manager.

    ``display_name`` (*string*)
      Friendly name for the collection. Unicode string, max 128 characters,
      no new lines (`\r` or `\n`).

    ``root_path`` (*string*)
      Root path for the collection, relative to the Storage Gateway. [Private]

    ``default_directory`` (*string*)
      Default directory when using the collection, relative to the root_path.

    ``public`` (*bool*)
      Flag indicating whether this collection is visible to other Globus users.

    ``force_encryption`` (*bool*)
      Flag indicating whether all data transfers to and from this collection
      are always encrypted.

    ``polices`` (*dict*)
      Connector specific collection policies. Most connectors do not use
      these. It is recommended that you use a CollectionPolicy helper classes
      (e.g. `GoogleCloudStorageCollectionPolicies` if you are using the
      Google Cloud Storage connector) to create these.

    ``organization`` (*string*)
      Organization that maintains the collection.

    ``department`` (*string*)
      Department within organization that runs the server(s).

    ``keywords`` (*list of strings*)
      List of search keywords for the endpoint

    ``description`` (*string*)
      A description of the endpoint.

    ``contact_email`` (*string*)
      Email address of the support contact for the collection.

    ``contact_info`` (*string*)
      Other non-email contact information for the collection, e.g. phone and
      mailing address.

    ``info_link``  (*string*)
      Link to a web page with more information about the endpoint.

    ``allow_guest_collections`` (*bool*)
      Flag indicating if this collection allows users to create guest
      collections on it. It may only be set to True on mapped collections.

    ``domain_name`` (*string*)
      Host name label for this collection (mapped only).

    ``sharing_restrict_paths`` (*dict*)
      Restrictions on which paths may be shared in guest
      collections related to this mapped collection [private].

    ``enable_https`` (*bool*)
      Boolean flag indicating whether this collection should support HTTPS or
      not.

    ``user_message`` (*str*)
      A message for clients to display to users when interacting with this
      collection.

    ``user_message_link`` (*str*)
      Link to additional messaging for clients to display to users when
      interacting with this endpoint, linked to an http or https URL.
    """

    def __init__(
        self,
        DATA_TYPE=None,
        storage_gateway_id=None,
        identity_id=None,
        collection_type=None,
        display_name=None,
        root_path=None,
        default_directory=None,
        policies=None,
        public=None,
        force_encryption=None,
        allow_guest_collections=None,
        disable_anonymous_writes=None,
        organization=None,
        department=None,
        keywords=None,
        description=None,
        contact_email=None,
        contact_info=None,
        info_link=None,
        domain_name=None,
        **kwargs
    ):
        if DATA_TYPE is None:
            if disable_anonymous_writes is not None:
                DATA_TYPE = "collection#1.5.0"
            elif "force_verify" in kwargs:
                DATA_TYPE = "collection#1.4.0"
            elif "sharing_users_allow" in kwargs or "sharing_users_deny" in kwargs:
                DATA_TYPE = "collection#1.2.0"
            elif (
                kwargs.get("enable_https") is not None
                or "user_message" in kwargs
                or "user_message_link" in kwargs
            ):
                DATA_TYPE = "collection#1.1.0"
            else:
                DATA_TYPE = "collection#1.0.0"

        _set_string_kwargs(
            self,
            DATA_TYPE=DATA_TYPE,
            storage_gateway_id=storage_gateway_id,
            identity_id=identity_id,
            collection_type=collection_type,
            display_name=display_name,
            root_path=root_path,
            default_directory=default_directory,
            organization=organization,
            department=department,
            description=description,
            contact_email=contact_email,
            contact_info=contact_info,
            info_link=info_link,
            domain_name=domain_name,
        )

        _set_bool_kwargs(
            self,
            allow_guest_collections=allow_guest_collections,
            force_encryption=force_encryption,
            public=public,
            disable_anonymous_writes=disable_anonymous_writes,
        )
        _set_raw_kwargs(self, keywords=keywords, policies=policies)

        self.update(kwargs)


class GoogleCloudStorageCollectionPolicies(dict):
    """
    Convenience class for constructing a Google Cloud Storage Collection Policy
    document to use as the `policies` parameter when creating a
    CollectionDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``project`` (*string*)
      Google Cloud Platform project ID that is used by this collection
    """

    def __init__(
        self,
        DATA_TYPE="google_cloud_storage_collection_policies#1.0.0",
        project=None,
        **kwargs
    ):
        self["DATA_TYPE"] = safe_stringify(DATA_TYPE)
        _set_string_kwargs(self, project=project)

        self.update(kwargs)


class POSIXCollectionPolicies(dict):
    """
    Convenience class for constructing a Posix Collection Policy
    document to use as the `policies` parameter when creating a
    CollectionDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``sharing_groups_allow`` (*list of strings*)
      POSIX groups which are allowed to create guest collections.

    ``sharing_groups_deny`` (*list of strings*)
      POSIX groups which are not allowed to create guest collections.
    """

    def __init__(self, DATA_TYPE="posix_collection_policies#1.1.0", **kwargs):
        self["DATA_TYPE"] = safe_stringify(DATA_TYPE)

        if "sharing_groups_allow" in kwargs:
            _set_string_list_kwargs(
                self, sharing_groups_allow=kwargs["sharing_groups_allow"]
            )
        if "sharing_groups_deny" in kwargs:
            _set_string_list_kwargs(
                self, sharing_groups_deny=kwargs["sharing_groups_deny"]
            )


class POSIXStagingCollectionPolicies(POSIXCollectionPolicies):
    """
    Convenience class for constructing a Posix Staging Collection Policy
    document to use as the `policies` parameter when creating a
    CollectionDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``sharing_groups_allow`` (*list of strings*)
      POSIX groups which are allowed to create guest collections.

    ``sharing_groups_deny`` (*list of strings*)
      POSIX groups which are not allowed to create guest collections.
    """

    def __init__(self, DATA_TYPE="posix_staging_collection_policies#1.0.0", **kwargs):
        super().__init__(DATA_TYPE=DATA_TYPE, **kwargs)


class RoleDocument(dict):
    """
    Convenience class for constructing a Role document
    to use as the `data` parameter to :meth:`create_role`

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``role`` (*string*)
      Role assigned to the principal. Known values are owner,
      administrator, access_manager, activity_manager,
      and activity_monitor

    ``principal`` (*string*)
      Auth identity or group id URN. Should be in the format
      urn:globus:auth:[identity|group]:{uuid of identity or group}

    ``collection`` (*string*)
      Collection ID for the collection the role will apply to.
      This value is omitted when creating an endpoint
      role or when creating role definitions when creating collections.
    """

    def __init__(
        self,
        DATA_TYPE="role#1.0.0",
        role=None,
        principal=None,
        collection=None,
        **kwargs
    ):
        _set_string_kwargs(
            self,
            DATA_TYPE=DATA_TYPE,
            role=role,
            principal=principal,
            collection=collection,
        )
        self.update(kwargs)


class NodeDocument(dict):
    """
    Convenience class for constructing a Node document to use as the `data`
    parameter to :meth:`node_update`

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.
    """

    def __init__(
        self,
        # For backwards compatibility
        DATA_TYPE="node#1.0.0",
        **kwargs
    ):
        if "data_interface" in kwargs:
            DATA_TYPE = "node#1.1.0"
        self["DATA_TYPE"] = DATA_TYPE
        self.update(kwargs)


class SharingPolicyDocument(dict):
    """
    Convenience class for constructing a SharingPolicy document to use as the
    `data` parameter to :meth:`create_sharing_policy`

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.
    """

    def __init__(
        self,
        DATA_TYPE="sharing_policy#1.0.0",
        users=None,
        read=None,
        read_write=None,
        none=None,
        **kwargs
    ):
        if users:
            self["users"] = users
        self["DATA_TYPE"] = DATA_TYPE
        self["sharing_restrict_paths"] = {
            "DATA_TYPE": "path_restrictions#1.0.0",
            "read": read if read else [],
            "read_write": read_write if read_write else [],
            "none": none if none else [],
        }

        self.update(kwargs)
