"""
This module is taken from the SDK ALPHA !

There's no guarantee that the code will be mainlined like this, and it might change
significantly before becoming part of the SDK.
However, it provides good utilities for paginated methods within an SDK Client.
"""


class PaginatorCollection:
    """
    A PaginatorCollection maps multiple methods of an SDK client to paginated variants.
    Given a method, client.foo , the collection will gain a function attribute `foo`
    (name matching is automatic) which returns a Paginator.

    So

    >>> pc = PaginatorCollection(client.foo)  # build it
    >>> paginator = pc.foo()  # returns a paginator
    >>> for page in paginator:  # a paginator is an iterable of pages (response objects)
    >>>     print(json.dumps(page.data))  # you can handle each response object in turn

    The expected usage is to build a `paginated` attribute of a client.

    That is, if `client` has two methods `foo` and `bar` which we want paginated,

    >>> client.paginated = PaginatorCollection(client.foo, client.bar)

    and that will let us call

    >>> client.paginated.foo()
    >>> client.paginated.bar()
    """

    def __init__(self, *simple_method_list, spec=None):
        if not spec:
            spec = {}

        self._paginator_map = {}

        for bound_method in simple_method_list:
            self._add_binding(MarkerPaginator, {}, bound_method)
        for paginator_class, spec in spec.items():
            for bound_method, paginator_args in spec.items():
                self._add_binding(paginator_class, paginator_args, bound_method)

    # This needs to be a separate callable to work correctly as a closure over loop
    # variables
    #
    # If you try to dynamically define functions in __init__ without wrapping them like
    # this, then you run into trouble with the fact that the resulting functions will
    # point at <locals> which aren't resolved until call time
    def _add_binding(self, paginator_class, paginator_args, bound_method):
        methodname = bound_method.__name__

        def paginated_method(*args, **kwargs):
            return paginator_class(
                bound_method, client_args=args, client_kwargs=kwargs, **paginator_args
            )

        self._paginator_map[methodname] = paginated_method

    def __getattr__(self, attrname):
        try:
            return self._paginator_map[attrname]
        except KeyError:
            pass
        raise AttributeError("No known paginator named '{}'".format(attrname))


class Paginator:
    """
    Base class for all paginators
    """


class MarkerPaginator(Paginator):
    """
    A paginator which uses `has_next_page` and `marker` from payloads, sets the `marker`
    query param to page.

    This is the default method for GCS pagination, so it's very simple.
    """

    def __init__(self, method, client_args, client_kwargs):
        """
        **Parameters**

        ``method``
          A bound method of an SDK client, used to generate a paginated variant

        ``client_args``
          Arguments to the underlying method which are passed when the paginator is
          instantiated. i.e. given ``client.paginated.foo(a, b, c=1)``, this will be
          ``[a, b]``. The paginator will pass these arguments to *each* call of the
          bound method as it pages.

        ``client_kwargs``
          Keyword arguments to the underlying method, like ``client_args`` above.
          ``client.paginated.foo(a, b, c=1)`` will pass this as ``{"c": 1}``. As with
          ``client_args``, it's passed to each paginated call.
        """
        self.method = method
        self.marker = None
        self.client_args = client_args
        self.client_kwargs = client_kwargs

    def __iter__(self):
        has_next_page = True
        while has_next_page:
            if self.marker:
                self.client_kwargs["marker"] = self.marker
            current_page = self.method(*self.client_args, **self.client_kwargs)
            yield current_page
            self.marker = current_page.get("marker")
            has_next_page = current_page.get("has_next_page", False)
