import click

import globus_sdk
from globus_sdk.exc import AuthAPIError

from ..config import (
    AUTH_CLIENT_DATA_CONFIG_KEY,
    internal_auth_client,
    internal_native_client,
    load_client_data,
    token_storage_adapter,
)

_RESCIND_HELP = """\
Rescinding Consents
-------------------
The logout command only revokes tokens that it can see in its storage.
If you are concerned that logout may have failed to revoke a token,
you may want to manually rescind the GCS v4 Migration consent on the
Manage Consents Page:

    https://auth.globus.org/consents
"""


_LOGOUT_EPILOG = """\
You are now successfully logged out of the GCS v4 Migration.
You may also want to logout of any browser session you have with Globus:

  https://auth.globus.org/v2/web/logout

Before attempting any further GCS v4 Migration commands, you will have to login
again using

  gcs-migration login
"""


@click.command("logout", short_help="Logout of the GCS v4 Migration")
@click.confirmation_option(
    prompt="Are you sure you want to logout?",
    help='Automatically say "yes" to all prompts',
)
def logout() -> None:
    """
    Logout of the GCS v4 Migration tool. Removes your Globus tokens from local
    storage, and revokes them so that they cannot be used anymore.
    """
    # try to get the user's preferred username from userinfo
    # if an API error is raised, they probably are not logged in

    adapter = token_storage_adapter()

    # we use a Native Client to prevent an invalid instance client
    # from preventing token revocation
    native_client = internal_native_client()

    services = ["auth.globus.org", "transfer.api.globus.org"]
    for service in services:
        # pull token data from the DB
        token_data = adapter.read_as_dict().get(service, {})

        # remove tokens from config and revoke them
        # also, track whether or not we should print the rescind help
        print_rescind_help = False
        for token_opt in ("access_token", "refresh_token"):
            # first lookup the token -- if not found we'll continue
            token = token_data.get(token_opt)
            if not token:
                click.echo(
                    (
                        'Warning: Found no token named "{}"! '
                        "Recommend rescinding consent"
                    ).format(token_opt)
                )
                print_rescind_help = True
                continue
            # token was found, so try to revoke it
            try:
                native_client.oauth2_revoke_token(token)
            # if we network error, revocation failed -- print message and abort so
            # that we can revoke later when the network is working
            except globus_sdk.NetworkError:
                click.echo(
                    (
                        "Failed to reach Globus to revoke tokens. "
                        "Because we cannot revoke these tokens, canceling "
                        "logout"
                    )
                )
                click.get_current_context().exit(1)

        # finally, we revoked, so it's safe to remove the tokens
        adapter.remove_tokens_for_resource_server(service)

    # delete the instance client if one exists
    client_data = load_client_data()

    if client_data:
        instance_client = internal_auth_client()
        try:
            instance_client.delete(
                "/v2/api/clients/{}".format(client_data["client_id"])
            )
        # if the client secret has been invalidated or the client has
        # already been removed, we continue on
        except AuthAPIError:
            pass

    # remove deleted client values
    adapter.remove_config(AUTH_CLIENT_DATA_CONFIG_KEY)

    # if print_rescind_help is true, we printed warnings above
    # so, jam out an extra newline as a separator
    click.echo(("\n" if print_rescind_help else "") + _LOGOUT_EPILOG)

    # if some token wasn't found in the config, it means its possible that the
    # config file was removed without logout
    # in that case, the user should rescind the CLI consent to invalidate any
    # potentially leaked refresh tokens, so print the help on that
    if print_rescind_help:
        click.echo(_RESCIND_HELP)
