"""
This module defines decorators which can be used to control access to various
GCS Manager API routes.
"""
from functools import partial, wraps
from typing import Any, Callable, List, Optional, Union

from flask import current_app, g
from globus.manager.api.models.storage_gateway import StorageGateway

import globus.manager.api.errors as errors
from .authentication_timeout_filter import AuthenticationTimeoutFilter
from .credential_owner import CredentialOwnerAuthorizer
from .identity_set import IdentitySet
from .public import PublicAuthorizer
from .role import (
    RoleAuthorizer,
    RoleSpec,
)
from .storage_gateway import StorageGatewayAuthorizer


def requires_one_role(
    fn: Callable = None,
    role_names: List[str] = None,
    require_session: Optional[bool] = None,
    authentication_timeout: Optional[Union[int, bool]] = False,
    require_mfa: Optional[bool] = None,
    storage_gateway_selector: Callable = None,
    collection_selector: Callable = None,
):
    """
    Decorate a route to enforce a policy requiring the identity set of the
    caller have access to at least one of the listed roles.  If any of the
    roles is granted to the caller's identity set, decorator calls `fn`.

    :param callable fn:
        Function to call if the authorization requirements are met.
    :param List[str] role_names:
        A list of role string that can be processed with the
        `globus.manager.api.authorization.role.RoleSpec` constructor.
    :require_session:
        If True, roles are only considered if the caller's identity with
        that role has authenticated within the current request's Globus Auth
        session. By default, sessions are not required. If a value for this
        parameter is not provided, it will be derived from the policies of the
        storage gateway returned by calling the `storage_gateway_selector`
        or the policies of the storage gateway associated with the collection
        returned from calling the `collection_selector`.
    :authentication_timeout:
        If set, roles are only considered if the caller's identity with
        that rolehas authenticated within this timeout (in minutes). If
        `require_session` is True, then this filters based on session
        authentications, otherwise it filters based on last_authentications
        from any client application. If a value for this
        parameter is not provided, it will be derived from the policies of the
        storage gateway returned by calling the `storage_gateway_selector`
        or the policies of the storage gateway associated with the collection
        returned from calling the `collection_selector`.
    :require_mfa:
        If True, roles are only considered if the caller's identity with
        that role has authenticated within the current request's Globus Auth
        session using multi-factor authentication. By default, multi-factor
        authentication is not required. If a value for this
        parameter is not provided, it will be derived from the policies of the
        storage gateway returned by calling the `storage_gateway_selector`
        or the policies of the storage gateway associated with the collection
        returned from calling the `collection_selector`.
    :param callable storage_gateway_selector:
        A callable which will be called to resolve a storage_gateway associated
        with the current request.
    :param callable collection_selector:
        A callable which will be called to resolve a collection associated with
        the current request. This is required to evaluate collection
        roles but is not needed for endpoint roles.

    :raises MissingRequiredRoleError:
        The desired role is not available to the caller.
    """
    if fn is None:
        return partial(
            requires_one_role,
            role_names=role_names,
            require_session=require_session,
            authentication_timeout=authentication_timeout,
            require_mfa=require_mfa,
            storage_gateway_selector=storage_gateway_selector,
            collection_selector=collection_selector,
        )

    @wraps(fn)
    def decorated_function(*args, **kwargs):
        collection = None
        role_specs = []
        if collection_selector:
            collection = collection_selector()
            if role_names:
                for role_name in role_names:
                    # Ignore collection-specific roles when no collection is
                    # selected
                    if collection is not None or not role_name.endswith(
                        ":{collection_id}"
                    ):
                        role_specs.append(RoleSpec(role_name, collection))
        else:
            for role_name in role_names:
                role_specs.append(RoleSpec(role_name))

        role_authorizer = RoleAuthorizer(role_specs)
        g.authorizer = role_authorizer
        try:
            g.auth_state
        except AttributeError:
            raise errors.UnauthorizedError()

        authorized_identities = role_authorizer.authorize_identities(g.auth_state)

        if authorized_identities:
            # Filter identities which have a required role based on the
            # required session and/or timeout information, if there is any.

            storage_gateway = None
            if storage_gateway_selector:
                # Override defaults if storage_gateway_selector is present
                storage_gateway = storage_gateway_selector()
                if not storage_gateway:
                    raise errors.NotFoundError()
            elif collection:
                # Override defaults if storage_gateway_selector is present
                storage_gateway = collection.storage_gateway

            if storage_gateway:
                current_app.logger.debug(
                    "authorizer using storage_gateway %s with require_session=%s, "
                    "authentication_timeout=%s and require_mfa=%s, passed in "
                    "require_session=%s, authentication_timeout=%s and require_mfa=%s",
                    storage_gateway,
                    storage_gateway.high_assurance,
                    storage_gateway.authentication_timeout_mins,
                    storage_gateway.require_mfa,
                    require_session,
                    authentication_timeout,
                    require_mfa,
                )

            if require_session is None:
                if storage_gateway:
                    authorizer_require_session = storage_gateway.high_assurance
                else:
                    authorizer_require_session = False
            else:
                authorizer_require_session = require_session

            if authentication_timeout is False:
                if storage_gateway:
                    authorizer_authentication_timeout = (
                        storage_gateway.authentication_timeout_mins
                    )
                else:
                    authorizer_authentication_timeout = None
            else:
                authorizer_authentication_timeout = authentication_timeout

            if require_mfa is None:
                if storage_gateway:
                    authorizer_require_mfa = storage_gateway.require_mfa
                else:
                    authorizer_require_mfa = False
            else:
                authorizer_require_mfa = require_mfa

            authentication_timeout_filter = AuthenticationTimeoutFilter(
                require_session=authorizer_require_session,
                authentication_timeout=authorizer_authentication_timeout,
                require_mfa=authorizer_require_mfa,
            )
            authorized_identities = authentication_timeout_filter(
                authorized_identities, g.auth_state
            )

            if authorized_identities:
                g.authorized_identities = authorized_identities
                return fn(*args, **kwargs)
            else:
                # Caller has a role, but it's not in session
                g.authorized_identities = []

                if authorized_identities.error_detail:
                    raise errors.AuthenticationTimeoutError(
                        detail=authorized_identities.error_detail
                    )
                else:
                    raise errors.MissingRequiredRoleError(roles=role_specs)
        else:
            # No role is available
            g.authorized_identities = []
            raise errors.MissingRequiredRoleError(roles=role_specs)

    return decorated_function


def requires_credential_owner(fn=None, credential_selector=None):
    """
    Decorate a route to enforce a policy requiring the identity set of the
    caller be the owner of a specific
    `globus.manager.api.models.UserCredential`. If an identity
    in the caller's identity set owns this credntial, this decorator calls
    `fn`.

    :param callable fn:
        Function to call if the authorization requirements are met.
    :param callable credential_selector:
        A callable which will be called to resolve a credential associated with
        the current request. This is required to evaluate collection
        roles but is not needed for endpoint roles.

    :raises NotResourceOwnerError:
        The credential is not owned by the coller.
    :raises InternalServerError:
        The credential_selector is misconfigured.
    """
    if fn is None:
        return partial(
            requires_credential_owner, credential_selector=credential_selector
        )

    @wraps(fn)
    def decorated_function(*args, **kwargs):
        g.authorized_identities = IdentitySet()
        if credential_selector:
            credential = credential_selector()
            if not credential:
                raise errors.NotFoundError()
            if credential:
                credential_authorizer = CredentialOwnerAuthorizer(credential)
                g.authorized_identities = credential_authorizer.authorize_identities(
                    g.auth_state
                )
                g.authorizer = credential_authorizer
                if g.authorized_identities:
                    return fn(*args, **kwargs)
            if g.authorized_identities.error_detail:
                raise errors.AuthorizationError(
                    detail=g.authorized_identities.error_detail
                )
            else:
                raise errors.NotResourceOwnerError(id=credential.id)
        else:
            raise errors.InternalServerError()

    return decorated_function


def requires_data_access_scope(fn: Callable = None):
    """
    Decorate a route to require the caller to have the `data_access` scope for
    the collection model stored in g.mapped_collection if it is not a high
    assurance collection. If the scope is not present and the collection is
    not high assurance, return a 401 error requiring new authorization with
    that scope as a dependent scope of the manage_collections scope.

    :param fn:
        The route handler to call after the scope check
    """
    if fn is None:
        return partial(requires_data_access_scope)

    @wraps(fn)
    def decorated_function(*args, **kwargs):
        collection = g.mapped_collection

        if not collection:
            raise errors.BadRequestError("Bad or missing mapped_collection_id")

        if not collection.high_assurance:
            data_access_scope_name = "https://auth.globus.org/scopes/{}/data_access".format(
                collection.id
            )
            manage_collections_scope_name = "urn:globus:auth:scope:{}:manage_collections".format(
                current_app.config["CLIENT_ID"]
            )

            scopes = [
                "{}[*{}]".format(manage_collections_scope_name, data_access_scope_name,)
            ]

            try:
                g.auth_state.dependent_token(
                    collection.id, required_scope=data_access_scope_name
                )
            except errors.MissingRequiredScopesError:
                raise errors.MissingRequiredScopesError(scopes=scopes)
        return fn(*args, **kwargs)

    return decorated_function


def requires_storage_gateway_policy(fn=None, storage_gateway_selector=None):
    """
    Decorate a route to enforce a policy requiring the identity set of the
    caller be allowed by the policies associated with a particular storage
    gateway. If an identity in the caller's identity set owns this credential,
    this decorator calls `fn`.

    :param callable fn:
        Function to call if the authorization requirements are met.
    :param callable storage_gateway_selector:
        A callable which will be called to resolve a storage_gateway associated
        with the current request. This must return a
        `globus.manager.api.models.StorageGateway` if the request can be
        resolved to one, or None otherwise.

    :raises BadRequestError:
        The storage_gateway_selector did not return a StorageGateway object.
    :raises ForbiddenError:
        The caller is not authorized by the Storage Gateway policies.
    """
    if fn is None:
        return partial(
            requires_storage_gateway_policy,
            storage_gateway_selector=storage_gateway_selector,
        )

    @wraps(fn)
    def decorated_function(*args, **kwargs):
        g.authorized_identities = []
        storage_gateway = None
        if storage_gateway_selector:
            storage_gateway = storage_gateway_selector()
        else:
            raise errors.InternalServerError()
        if storage_gateway:
            storage_gateway_authorizer = StorageGatewayAuthorizer(storage_gateway)
            g.authorizer = storage_gateway_authorizer
            authorized_identities = storage_gateway_authorizer.authorize_identities(
                g.auth_state
            )

            if authorized_identities:
                g.authorized_identities = authorized_identities
                return fn(*args, **kwargs)
            else:
                raise errors.AuthorizationError(
                    detail=authorized_identities.error_detail
                )
        else:
            raise errors.BadRequestError()

    return decorated_function


def requires_endpoint_visibility(fn=None):
    """
    Decorate a route to enforce a policy requiring the endpoint to be visible
    to an identity in the caller's identity_set.  An endpoint
    is visible to the caller if any of the following are true:

    -   The endpoint document's `public` property is True
    -   The caller has any role assignment for the endpoint or any
        collection on the endpoint
    -   The caller has an ACL granting access to any guest collection on
        this endpoint.
    -   The caller has an identity allowed by any storage gateway policy
        on this endpoint.

    :param callable fn:
        Function to call if the authorization requirements are met.

    :raises ForbiddenError:
        The caller is not authorized to view the endpoint.
    """

    @wraps(fn)
    def decorated_function(*args, **kwargs):
        public_authorizer = PublicAuthorizer()
        authorized = public_authorizer.authorize_identities(g.auth_state)

        if authorized:
            g.authorizer = public_authorizer
        else:
            role_authorizer = RoleAuthorizer(
                [
                    "endpoint:owner",
                    "endpoint:administrator",
                    "endpoint:activity_manager",
                    "endpoint:activity_monitor",
                    "collection:administrator:*",
                    "collection:activity_manager:*",
                    "collection:activity_monitor:*",
                    "collection:access_manager:*",
                ]
            )

            authorized = role_authorizer.authorize_identities(g.auth_state)
            if authorized:
                g.authorized = role_authorizer

        # TODO: ACL authorizer check
        if not authorized:
            for sg in StorageGateway.query.all():
                sg_authorizer = StorageGatewayAuthorizer(sg)
                authorized = sg_authorizer.authorize_identities(g.auth_state)
                if authorized:
                    g.authorizer = sg_authorizer
                    break

        if authorized:
            g.authorized_identities = authorized
            return fn(*args, **kwargs)
        else:
            g.authorized_identities = []
            if authorized.error_detail:
                raise errors.AuthorizationError(
                    error_description="This endpoint is not "
                    "visible to any of your identities",
                    detail=authorized.error_detail,
                )

    return decorated_function


def requires_subscription(fn: Callable = None, level: Optional[str] = None) -> Any:
    """
    Decorate a route to require a subscription for the route to be
    considered accessible. If no subscription is present on the endpoint,
    raise the subscription error. This is intended to be used before
    role based authorization checks in order to make the errors more
    understandable.

    :param fn:
        The route handler to call after the subscription check.
    :param level:
        The minimum subscription level to use. If this is not supplied by the
        caller, `starter` is used.
    """
    if fn is None:
        return partial(requires_subscription, level=level)

    if level is None:
        level = "starter"

    @wraps(fn)
    def decorated_function(*args, **kwargs):
        if not current_app.subscription and level != "basic":
            raise errors.SubscriptionRequiredError(subscription_level=level)
        return fn(*args, **kwargs)

    return decorated_function
