import re
from typing import Any, List, Mapping, Optional
import uuid

from flask import (
    Blueprint,
    current_app,
    g,
    jsonify,
    request,
    url_for,
    Response,
)
import marshmallow as ma

import globus_sdk

from globus.manager.api.authentication import authenticated
from globus.manager.api.authorization import (
    AuthenticationTimeoutFilter,
    EndpointRoleAuthorizer,
    IdentitySet,
    PublicAuthorizer,
    RoleAuthorizer,
    RoleSpec,
    StorageGatewayAuthorizer,
)
from globus.manager.api.authorization.decorators import (
    requires_data_access_scope,
    requires_one_role,
    requires_storage_gateway_policy,
    requires_subscription,
)
from globus.manager.api.blueprints.decorators import (
    Base62Marker,
    expects_json,
    paginated,
)
from globus.manager.api.blueprints.utils import (
    check_subscription,
    create_object,
    delete_object,
    patch_object,
    put_object,
    update_object,
    validate_unique_custom_domain,
)
import globus.manager.constants as c
from globus.manager.api.models.result import Result
import globus.manager.api.errors as errors
from globus.manager.database import db
import globus.manager.api.models as m
import globus.manager.api.schemas.api as api_schema


blueprint = Blueprint("collections", __name__, url_prefix="/api/")


@blueprint.route("/collections", methods=["GET"])
@paginated
@authenticated
def collections_get():
    mapped_collection_id = request.args.get("mapped_collection_id", None)
    filters = _parse_filters(request.args.get("filter"))
    include_private_policies = "private_policies" in request.args.get(
        "include", ""
    ).split(",")

    marker = g.marker
    page_size = g.page_size
    next_page_token = None

    filter_managed_by_me = "managed_by_me" in filters

    # We'll check endpoint-level access (public and roles) before
    # iterating on the collection and apply the results to every collection
    if filter_managed_by_me:
        # If we are filtering by role, we don't care if the endpoint is public
        # and don't bother checking here.
        endpoint_is_public = False
    else:
        endpoint_is_public = PublicAuthorizer().authorize_identities(g.auth_state)

    # If the caller has an endpoint role, they can see all collections
    caller_has_endpoint_role = EndpointRoleAuthorizer.authorize_identities(g.auth_state)

    # The query can support the non-role based filters, but roles are
    # more complicated because of group permissions
    query = m.Collection.query.filter(~m.Collection.deleted)  # type: ignore
    filter_created_by_me = False
    for f in filters:
        if f == "mapped_collections":
            query = query.filter(m.Collection.is_mapped_collection)
        elif f == "guest_collections":
            query = query.filter(m.Collection.is_guest_collection)
        elif f == "created_by_me":
            filter_created_by_me = True

    if marker:
        query = query.filter(m.ConfigurationObject.id >= marker.marker)

    if mapped_collection_id:
        query = query.filter(
            m.Collection.collection_type == c.CollectionType.guest.name,
            m.Collection.mapped_collection_id == str(mapped_collection_id),
        )
    query = query.order_by(m.ConfigurationObject.id)

    results = []
    for col in query:
        if filter_created_by_me:
            if col.identity_id not in g.auth_state.identity_set:
                continue
            else:
                current_app.logger.debug(
                    "I think %s is created by me because %s is in %s",
                    col,
                    col.identity_id,
                    g.auth_state.identity_set,
                )

        if not _check_collection_visibility(
            col,
            endpoint_is_public=endpoint_is_public,
            caller_has_endpoint_role=caller_has_endpoint_role,
            filter_managed_by_me=filter_managed_by_me,
        ):
            continue

        if len(results) < page_size:
            serialized_collection = None
            if include_private_policies:
                admin_ids = _caller_is_collection_admin(col)
                if admin_ids:
                    serialized_collection = api_schema.collection_private_schema.dump(
                        col
                    )
            if not serialized_collection:
                serialized_collection = api_schema.collection_schema.dump(col)

            results.append(serialized_collection)
        else:
            next_page_token = Base62Marker(marker_data=col.id)
            break

    if (
        not endpoint_is_public
        and len(results) == 0
        and not mapped_collection_id
        and not filters
        and not caller_has_endpoint_role
    ):
        raise errors.ForbiddenError()

    res = api_schema.result_schema.dump(Result(data=results, marker=next_page_token))

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


@blueprint.route("/collections/<uuid:collection_id>", methods=["GET"])
@authenticated
def collections_get_by_id(collection_id: uuid.UUID = None):
    include_private_policies = "private_policies" in request.args.get(
        "include", ""
    ).split(",")
    query = m.Collection.query.filter(
        m.Collection.id == str(collection_id), ~m.Collection.deleted  # type: ignore
    )
    col = query.one_or_none()

    if not col:
        raise errors.NotFoundError()

    endpoint_is_public = PublicAuthorizer().authorize_identities(g.auth_state)

    # If the endpoint is public, presence of the role doesn't matter
    caller_has_endpoint_role = False
    if not endpoint_is_public:
        caller_has_endpoint_role = bool(
            EndpointRoleAuthorizer.authorize_identities(g.auth_state)
        )

    if not _check_collection_visibility(
        col,
        endpoint_is_public=endpoint_is_public,
        caller_has_endpoint_role=caller_has_endpoint_role,
    ):
        if endpoint_is_public:
            raise errors.NotFoundError()
        else:
            raise errors.ForbiddenError()

    res = None
    if include_private_policies:
        authorized_admins = _caller_is_collection_admin(col)
        if authorized_admins:
            res = api_schema.result_schema.dump(
                Result(data=[api_schema.collection_private_schema.dump(col)])
            )
        else:
            detail = None
            current_app.logger.debug("details=%s", authorized_admins.error_details)
            if authorized_admins.error_details:
                detail = authorized_admins.error_details[0]
            raise errors.AuthorizationError(detail=detail)
    else:
        res = api_schema.result_schema.dump(
            Result(data=[api_schema.collection_schema.dump(col)])
        )

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


def _parse_filters(filters=None):
    """
    Parse the filters query parameter for GET /collections

    :param filters: The query parameter string.

    :returns:
        A list of validated filter names.
    :raises errors.BadRequestError:
        One or more filter name is not supported by this API.
    """
    if filters:
        filters = [f for f in filters.split(",") if f]
        allowed_filters = [
            "mapped_collections",
            "guest_collections",
            "managed_by_me",
            "created_by_me",
        ]

        bad_filters = [f for f in filters if f not in allowed_filters]
        if bad_filters:
            current_app.logger.error("filters='%s'", filters)
            raise errors.BadRequestError(
                "Unsupported filter: {}".format(",".join(bad_filters))
            )
    return filters or []


def _check_collection_visibility(
    collection,
    endpoint_is_public=False,
    caller_has_endpoint_role=False,
    filter_managed_by_me=False,
):
    """
    Perform collection visibility access checks based on the caller's
    identity set, collection, and endpoint policies.

    :param collection:
        The collection model to check for visibility.
    :param endpoint_is_public:
        True if the Endpoint.public property is True.
    :param caller_has_endpoint_role:
        True if the caller has any endpoint role.
    :param filter_managed_by_me:
        If true, the collection is visible only if the caller has an
        endpoint or collection role.

    :returns:
        Returns True if the collection is visible, False otherwise.
    """
    visible = caller_has_endpoint_role
    if visible:
        current_app.logger.debug("%s visible: caller has endpoint role", collection)

    if not visible and collection.public and not filter_managed_by_me:
        current_app.logger.debug("%s visible: collection is public", collection)
        visible = True

    if not visible:
        current_app.logger.debug("Checking for role granting access to %s", collection)

        collection_roles = [
            "collection:administrator:{collection_id}",
            "collection:activity_manager:{collection_id}",
            "collection:activity_monitor:{collection_id}",
            "collection:access_manager:{collection_id}",
        ]
        role_specs = [RoleSpec(r, collection) for r in collection_roles]
        collection_role_authorizer = RoleAuthorizer(role_specs)
        caller_has_collection_role = collection_role_authorizer.authorize_identities(
            g.auth_state
        )
        if bool(caller_has_collection_role):
            current_app.logger.debug(
                "%s visible: %s has a role on it",
                collection,
                caller_has_collection_role,
            )
            visible = True

    if not visible and not filter_managed_by_me:
        # TODO: ACL authorizer
        pass

    if not visible and not filter_managed_by_me:
        sg_authorizer = StorageGatewayAuthorizer(collection.storage_gateway)
        sg_authorized_identities = sg_authorizer.authorize_identities(g.auth_state)

        if bool(sg_authorized_identities):
            current_app.logger.debug(
                "%s visible: %s policies", collection, collection.storage_gateway
            )
            visible = True

    return visible


def _caller_is_collection_admin(collection) -> IdentitySet:
    """
    Check that the caller's identity set includes an identity with a
    collection:administrator role for this collection authenticated within the
    authentication_timeout_mins for the storage gateway this collection
    was created on.

    :param collection:
        The collection model to check for access to.
    :returns:
        This function returns an authorized IdentitySet object, which
        may be interpreted as a bool or checked for error details
    """
    current_app.logger.debug(
        "Checking whether caller is endpoint owner or collection_admin for %s",
        collection.id,
    )
    if not current_app.subscription:
        authorizer = RoleAuthorizer("endpoint:owner")
    else:
        authorizer = RoleAuthorizer(
            [
                RoleSpec("collection:administrator:{collection_id}", collection),
                "endpoint:owner",
            ]
        )

    idset = authorizer.authorize_identities(g.auth_state)

    current_app.logger.debug("identities to consider: %s", idset)

    # Enforce authorization timeouts
    idset_filter = AuthenticationTimeoutFilter(
        require_session=collection.storage_gateway.high_assurance,
        authentication_timeout=(collection.storage_gateway.authentication_timeout_mins),
        require_mfa=collection.require_mfa,
    )

    idset = idset_filter(idset, g.auth_state)

    current_app.logger.debug("result is %s", idset)
    return idset


@blueprint.route("/collections", methods=["POST"])
@authenticated
@expects_json(schema=api_schema.collection_schema)
def collections_post():
    collection_doc = g.validated_data
    with current_app.appsync_client.pause_notifications():
        collection_type = collection_doc["collection_type"]
        assert collection_type in c.CollectionType
        if collection_type == c.CollectionType.mapped:
            collection = collections_post_mapped()
        elif collection_type == c.CollectionType.guest:
            collection = collections_post_guest()

    if collection.owner_role:
        create_object(collection.owner_role)
        g.created_objects.append(str(collection.owner_role.id))

    res = api_schema.result_schema.dump(
        Result(
            http_response_code=201,
            message="Created collection {}".format(collection.id),
            data=[api_schema.collection_schema.dump(collection)],
        )
    )

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    response.headers["Location"] = url_for(
        ".collections_get_by_id",
        collection_id=collection.id,
        _external=True,
        _scheme="https",
    )

    return response


def allocate_domain(collection_type, mapped_collection=None) -> str:
    """
    Allocate a new domain name for a collection and update endpoint state in
    appsync.

    :param collection_type:
        Enum value indicating whether this is a mapped or guest collection.
    :param mapped_collection:
        Mapped collection object which this collection is being
        created on, if this is a guest collection; otherwise None.

    :returns:
        Returns a string containing the fully-qualified name to use for
        this collection
    """

    def update_domain_counter(endpoint, collection):
        nonlocal mapped_collection

        collection["domain_name"] = endpoint.allocate_domain_name(
            collection["collection_type"], mapped_collection
        )

    collection = {"collection_type": collection_type.name}
    endpoint = m.Endpoint.query.one()
    update_object(endpoint, update_domain_counter, collection)
    return collection["domain_name"]


def validate_custom_domain_label(domain_name: str, collection_id: str = None) -> str:
    """
    Ensure that the custom domain name for a mapped collection
    is valid, according to the following rules:

    -   The domain must not start with "g-" or "m-"
    -   If it contains a ".", the entire domain must be a child
        of the endpoint's wildcard domain
    -   The domain must not be used by any other collection.

    :param domain_name:
        Client-requested domain name.
    :param collection_id:
        ID of the collection we are updating. If None, then assume this
        is a create call. If we are updating, we don't error
        if the name is in use by itself.

    :returns:
        Returns the fully qualified domain name.

    :raises errors.ValidationError:
        The requested domain is not allowed because it does not match
        the required format.
    :raises errors.ConflictError:
        The requested domain is not allowed because it conflicts with
        an existing collection.
    """
    # Limited support for custom domains: must not start with g- or m-
    # and must be subdomains of the wildcard domain
    domain = None  # type: Optional[str]
    if "." in domain_name:
        label, domain = domain_name.lower().split(".", 1)
    else:
        label = domain_name
        domain = None

    if label == "" or not re.match(r"^(?![0-9]+$)(?!-)[a-zA-Z0-9-]{,63}(?<!-)$", label):
        raise errors.ValidationError("Invalid domain_name value")

    if re.match(r"^(m-|g-)", label):
        raise errors.ValidationError("Invalid domain_name value")

    endpoint_domain = m.Endpoint.query.one().collection_subdomain
    if domain:
        if domain != endpoint_domain:
            raise errors.ValidationError("Invalid domain_name value")

    new_domain_name = "{}.{}".format(label, endpoint_domain)

    in_use = m.Collection.query.filter(
        m.Collection.domain_name == new_domain_name
    ).one_or_none()

    if in_use and ((not collection_id) or in_use.id != collection_id):
        raise errors.ConflictError("domain_name already in use")

    return new_domain_name


def identity_data_for_requested_identity_id(
    identity_id: Optional[str], authorized_identities: List[Mapping[str, Any]]
) -> Mapping[str, Any]:
    """
    Determine which identity data in authorized_identities to use when
    creating a collection. If the caller has provided a requested `identity_id`,
    then return the identity from `authorized_identities` has that id;
    otherwise, choose the first identity in the `authorized_identities` list.

    :param identity_id:
        Optional string id of the requested identity. If this is not None, it
        must match the 'id' property of one of the elements of
        `authorized_identities`.
    :param authorized_identities:
        A list of identity_data dicts.

    :return:
        The member of `authorized_identities` whose `id` property
        matches the `identity_id`, or the first member of `authorized_identities`
        if `identity_id` is None.

    :raises errors.IdNotInIdentitySetError:
        No member of `authorized_identities` matches the requested
        `identity_id`.
    """
    if identity_id:
        authorized = [i for i in authorized_identities if i["id"] == identity_id]
        if not authorized:
            raise errors.IdNotInIdentitySetError(id=identity_id)
        return authorized[0]
    else:
        return g.authorized_identities[0]


def get_storage_gateway_from_collection_doc():
    storage_gateway_id = g.validated_data.get("storage_gateway_id")
    user_credential_id = g.validated_data.get("user_credential_id")
    mapped_collection_id = g.validated_data.get("mapped_collection_id")
    storage_gateway = None
    user_credential = None
    mapped_collection = None

    if storage_gateway_id:
        storage_gateway = m.StorageGateway.query.filter(
            ~m.StorageGateway.deleted, m.StorageGateway.id == str(storage_gateway_id)
        ).one_or_none()

    if user_credential_id:
        user_credential = m.UserCredential.query.filter(
            ~m.UserCredential.deleted, m.UserCredential.id == str(user_credential_id)
        ).one_or_none()
        if user_credential:
            storage_gateway = user_credential.storage_gateway

    if mapped_collection_id:
        mapped_collection = m.Collection.query.filter(
            m.Collection.id == str(mapped_collection_id)
        ).one_or_none()
        if mapped_collection:
            storage_gateway = mapped_collection.storage_gateway

    if storage_gateway is None:
        raise errors.BadRequestError("Invalid storage_gateway_id")

    g.storage_gateway = storage_gateway
    g.user_credential = user_credential
    g.mapped_collection = mapped_collection

    return storage_gateway


@requires_storage_gateway_policy(
    storage_gateway_selector=get_storage_gateway_from_collection_doc
)
@requires_subscription
@requires_data_access_scope
def collections_post_guest():
    collection_doc = g.validated_data

    user_credential = None
    storage_gateway = None
    identity_id = None  # type: Optional[uuid.UUID]
    domain_name = None
    endpoint = m.Endpoint.query.one()

    # Collection documents may have
    # user_credential_id:
    #   User credential associated with a guest collection. This must not
    #   be present in a mapped collection
    # identity_id:
    #   Auth identity of the creator of the collection. This can be guessed
    #   from either the user_credential_id or first entry that passes the
    #   identity mapping authorization step, or may be explicitly set.
    #
    # Whatever is present (if allowed for the collection type) must be
    # consistent:
    #
    # After resolving the storage_gateway_id and user_credential_id to
    # objects, user_credential.storage_gateway_id must equal
    # storage_gateway.id
    #
    # If identity_id is present it must be in the caller's identity set,
    # and in the list of authorized identities based on policies.
    storage_gateway = g.storage_gateway

    # If the caller has requested a specific user_credential_id be used,
    # verify that it exists
    user_credential_id = collection_doc.get("user_credential_id")
    if user_credential_id:
        user_credential = g.user_credential
        if user_credential is None:
            raise errors.BadRequestError("Invalid user_credential_id")
        elif not user_credential.provisioned:
            raise errors.InvalidCredentialError(
                "Your credential requires some initial setup.",
                user_credential_id=user_credential.id,
            )
        elif user_credential.invalid:
            raise errors.InvalidCredentialError(user_credential_id=user_credential.id)

    # If the caller has not requested a specific identity_id be used,
    # default to that of the user_credential if known
    identity_id = collection_doc.get("identity_id")
    if not identity_id and user_credential:
        identity_id = user_credential.identity_id

    # Check that user_credential is associated with the requested storage
    # gateway
    if user_credential and storage_gateway:
        if user_credential.storage_gateway_id != storage_gateway.id:
            raise errors.BadRequestError(
                "User Credential {} not associated with Storage Gateway {}".format(
                    user_credential.id, storage_gateway.id
                )
            )

    # Check that user_credential is associated with the requested
    # user_credential
    if identity_id and user_credential:
        if user_credential.identity_id != str(identity_id):
            raise errors.BadRequestError(
                "User Credential {} not associated with identity {}".format(
                    user_credential.id, identity_id
                )
            )

    check_subscription(
        storage_gateway.connector,
        high_assurance=storage_gateway.high_assurance,
        collection_type=collection_doc["collection_type"],
    )

    if not g.mapped_collection:
        raise errors.BadRequestError("Invalid mapped_collection_id")

    if not g.mapped_collection.allow_guest_collections:
        raise errors.ForbiddenError(
            "Collection {} does not allow guest collections".format(
                g.mapped_collection.id
            )
        )

    authorized_identity = identity_data_for_requested_identity_id(
        identity_id, g.authorized_identities
    )

    domain_name = allocate_domain(
        collection_doc["collection_type"], g.mapped_collection
    )

    # Create collection
    try:
        collection = m.Collection.from_collections_post(
            storage_gateway=storage_gateway,
            mapped_collection=g.mapped_collection,
            user_credential=user_credential,
            endpoint=endpoint,
            domain_name=domain_name,
            authorized_identity=authorized_identity,
            doc=g.validated_data,
        )
    except ValueError as v:
        raise errors.ValidationError(str(v) or None)

    create_object(collection)
    g.created_objects.append(collection.id)

    return collection


@requires_one_role(
    role_names=["endpoint:owner", "endpoint:administrator"],
    storage_gateway_selector=get_storage_gateway_from_collection_doc,
)
def collections_post_mapped():
    collection_doc = g.validated_data
    storage_gateway = g.storage_gateway
    endpoint = m.Endpoint.query.one()

    if collection_doc.get("allow_guest_collections", False):
        check_collection_type = c.CollectionType.guest
    else:
        check_collection_type = collection_doc["collection_type"]

    # Check for HA and connector support in the endpoint's subscription
    check_subscription(
        storage_gateway.connector,
        high_assurance=storage_gateway.high_assurance,
        collection_type=check_collection_type,
    )

    authorized_identity = identity_data_for_requested_identity_id(
        collection_doc.get("identity_id"), g.authorized_identities
    )

    if "domain_name" not in collection_doc:
        domain_name = allocate_domain(collection_doc["collection_type"])
    else:
        domain_name = validate_custom_domain_label(collection_doc["domain_name"])

    # Create collection
    try:
        collection = m.Collection.from_collections_post(
            storage_gateway=storage_gateway,
            endpoint=endpoint,
            domain_name=domain_name,
            authorized_identity=authorized_identity,
            doc=g.validated_data,
        )
    except ValueError as v:
        raise errors.ValidationError(str(v) or None)

    create_object(collection)
    g.created_objects.append(collection.id)

    return collection


def _update_collection(collection, collection_doc, update_func):
    """
    Common routine for put and patch for collections.
    """
    if collection is None:
        raise errors.NotFoundError()

    if collection.is_guest_collection:
        for field in ["user_message", "user_message_link"]:
            if field in collection_doc:
                exc = ma.ValidationError(
                    "Invalid field for guest collection.", field_name=field
                )
                raise errors.ValidationError(exc=exc)

    current_app.logger.debug(
        "Updating {} {} with {}".format(
            collection, collection.configuration_hash, collection_doc
        )
    )

    if collection_doc.get("allow_guest_collections", False):
        check_subscription(
            collection.storage_gateway.connector,
            high_assurance=collection.storage_gateway.high_assurance,
            collection_type=c.CollectionType.guest,
        )

    with current_app.appsync_client.pause_notifications():
        update_guests_transfer_docs = False
        guest_patch_doc = {}

        if (
            current_app.subscription
            and collection.is_mapped_collection
            and "enable_https" in collection_doc
            and collection.enable_https != collection_doc["enable_https"]
        ):
            update_guests_transfer_docs = True

        if (
            "user_message" in collection_doc
            and getattr(collection, "user_message", None)
            != collection_doc["user_message"]
        ):
            update_guests_transfer_docs = True

        if (
            "user_message_link" in collection_doc
            and getattr(collection, "user_message_link", None)
            != collection_doc["user_message_link"]
        ):
            update_guests_transfer_docs = True

        if (
            collection.is_mapped_collection
            and collection_doc.get("force_encryption", None) is not None
            and collection_doc["force_encryption"] != collection.force_encryption
        ):
            update_guests_transfer_docs = True
            if not collection_doc["force_encryption"]:
                guest_patch_doc["force_encryption"] = False

        if (
            collection.is_mapped_collection
            and collection_doc.get("force_verify", None) is not None
            and collection_doc["force_verify"] != collection.force_verify
        ):
            update_guests_transfer_docs = True
            if collection_doc["force_verify"]:
                guest_patch_doc["disable_verify"] = False
            else:
                guest_patch_doc["force_verify"] = False

        if (
            collection.is_mapped_collection
            and collection_doc.get("disable_verify", None) is not None
            and collection_doc["disable_verify"] != collection.disable_verify
        ):
            update_guests_transfer_docs = True
            if collection_doc["disable_verify"]:
                guest_patch_doc["force_verify"] = False
            else:
                guest_patch_doc["disable_verify"] = False

        if "domain_name" in collection_doc:
            if collection.is_mapped_collection:
                if (
                    collection.has_domain
                    and collection_doc["domain_name"] != collection.domain.domain_name
                ):
                    raise errors.ValidationError(
                        "Unable to change domain_name on a collection with a custom domain"
                    )
                domain_name = validate_custom_domain_label(
                    collection_doc["domain_name"], collection_id=collection.id
                )
                # The above returns fqdn, where input may be partial
                collection_doc["domain_name"] = domain_name
            elif collection.domain_name != collection_doc["domain_name"]:
                raise errors.ValidationError(
                    "Unable to change domain_name on a guest collection"
                )

        update_func(collection, collection_doc)

        if update_guests_transfer_docs:
            for guest in m.Collection.query.filter(
                ~m.Collection.deleted,
                m.Collection.is_guest_collection,
                m.Collection.mapped_collection_id == collection.id,
            ):
                try:
                    if guest_patch_doc:
                        # We explicitly update the guest model which requires a model+appsync update.
                        # patch() will call update_transfer_object()
                        patch_object(
                            guest,
                            {
                                "DATA_TYPE": collection_doc["DATA_TYPE"],
                                **guest_patch_doc,
                            },
                        )
                    else:
                        guest.update_transfer_endpoint()
                except globus_sdk.TransferAPIError:
                    pass  # Already calls log_transfer_exception(e)

        data = []
        data.append(api_schema.collection_schema.dump(collection))
        db.session.commit()

        res = api_schema.result_schema.dump(
            Result(message="Updated collection {}".format(collection.id), data=data)
        )

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    g.updated_objects.append(collection.id)

    return response


def get_collection_id_from_path():
    collection_id = request.view_args.get("collection_id")
    collection = None
    collection = m.Collection.query.filter(
        ~m.Collection.deleted, m.Collection.id == str(collection_id)
    ).one_or_none()

    g.collection = collection
    return collection


@blueprint.route("/collections/<uuid:collection_id>", methods=["PATCH"])
@authenticated
@requires_one_role(
    role_names=[
        "endpoint:owner",
        "endpoint:administrator",
        "collection:administrator:{collection_id}",
    ],
    collection_selector=get_collection_id_from_path,
)
@expects_json(schema=api_schema.collection_patch_schema)
def collections_patch_by_id(collection_id):
    return _update_collection(g.collection, g.validated_data, patch_object)


@blueprint.route("/collections/<uuid:collection_id>", methods=["PUT"])
@authenticated
@requires_one_role(
    role_names=[
        "endpoint:owner",
        "endpoint:administrator",
        "collection:administrator:{collection_id}",
    ],
    collection_selector=get_collection_id_from_path,
)
@expects_json(schema=api_schema.collection_schema)
def collections_put_by_id(collection_id):
    return _update_collection(g.collection, g.validated_data, put_object)


@blueprint.route("/collections/<uuid:collection_id>", methods=["DELETE"])
@authenticated
@requires_one_role(
    role_names=[
        "endpoint:owner",
        "endpoint:administrator",
        "collection:administrator:{collection_id}",
    ],
    collection_selector=get_collection_id_from_path,
)
def collections_delete_by_id(collection_id):
    if g.collection is None:
        raise errors.NotFoundError()

    collection = g.collection

    with current_app.appsync_client.pause_notifications():
        # (soft) Delete the collection
        delete_object(collection, update=True)

        # (soft) Delete related guest collections
        for gc in m.Collection.query.filter(
            ~m.Collection.deleted,
            m.Collection.mapped_collection_id == collection.id,
            m.Collection.is_guest_collection,
        ):
            gcid = gc.id
            delete_object(gc, update=True)
            g.deleted_objects.append(gcid)

            # Delete the roles for those
            for r in m.Role.query.filter(m.Role.resource == gc.id):
                rid = r.id
                if r.primary_role:
                    r.primary_role = False
                    r.refresh_configuration_hash()
                delete_object(r, collection_deleted=True)
                g.deleted_objects.append(rid)

        # Delete appsync representations of the role objects
        # for the mapped collection
        for r in m.Role.query.filter(m.Role.resource == collection.id):
            if r.primary_role:
                r.primary_role = False
                r.refresh_configuration_hash()
            rid = r.id
            delete_object(r, collection_deleted=True)
            g.deleted_objects.append(rid)

        # Delete sharing policies on the collection
        sharing_policy_ids = [
            sp.id
            for sp in m.SharingPolicy.query.filter(
                m.SharingPolicy.collection_id == collection.id
            )
        ]

        if sharing_policy_ids:

            def delete_policy(collection):
                for sharing_policy_id in sharing_policy_ids:
                    collection.delete_sharing_policy(sharing_policy_id)
                    g.deleted_objects.append(sharing_policy_id)

            update_object(g.collection, delete_policy)

    res = api_schema.result_schema.dump(
        Result(message="Deleted collection {}".format(str(collection_id)))
    )
    g.deleted_objects.append(collection.id)

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


@blueprint.route("/collections/<uuid:collection_id>/domain", methods=["PUT"])
@authenticated
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
@expects_json(schema=api_schema.domain_schema)
def collection_domain_put(collection_id: uuid.UUID) -> Response:
    """
    Register a new custom domain and certificate to to be used to serve
    this collection.

    The domain is used for the collection. If this is a mapped collection
    and the `wildcard` property is set to true, then all all guest collections
    associated with this collection that do not have a custom domain will be
    updated to use subdomains of that domain. Otherwise, only this collection
    will use that domain.

    This requires an `administrator` role on the Endpoint
    """
    with current_app.appsync_client.pause_notifications():
        collection = m.Collection.query.filter(
            ~m.Collection.deleted,  # type: ignore
            m.Collection.id == str(collection_id),
        ).one_or_none()  # type: Optional[m.Collection]

        if collection is None:
            raise errors.NotFoundError("Collection not found")

        old_domain_name = None  # type: Optional[str]
        try:
            old_domain = collection.domain
            old_domain_name = old_domain.domain_name
            old_wildcard = bool(old_domain and old_domain.wildcard)
        except AttributeError:
            old_wildcard = False

        if g.validated_data["domain_name"] != collection.domain_name:
            validate_unique_custom_domain(g.validated_data["domain_name"])

        update_object(collection, m.Collection.put_domain, g.validated_data)
        g.updated_objects.append(str(collection.id))

        if (
            old_wildcard or collection.domain.wildcard
        ) and old_domain_name != collection.domain.domain_name:
            _update_guest_collections_without_custom_domains(collection)

    res = api_schema.result_schema.dump(Result(message="Updated collection"))
    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


@blueprint.route("/collections/<uuid:collection_id>/domain", methods=["DELETE"])
@authenticated
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
def collection_domain_delete(collection_id: uuid.UUID) -> Response:
    """
    Delete the custom collection domain.

    If this is a mapped collection, this will cause the collection to revert to
    a subdomain of the endpoint's domain (if it is wildcard domain) or a
    subdomain of the endpoint's data.globus.org domain. If this mapped collection
    has a wildcard domain when this is called, then all guest collections without
    custom domains will have their domains changed as well.

    If this is a guest collection, and the mapped collection it was created from
    has a custom wildcard domain, then this collection will become a subdomain
    of that domain; otherwise it will revert to a subdomain of either the
    endpoint's domain (if it is a wildcard domain) or a subdomain of the endpoint's
    data.globus.org domain.

    This requires an `administrator` role on the Endpoint.
    """
    with current_app.appsync_client.pause_notifications():
        collection = m.Collection.query.filter(
            m.Collection.id == str(collection_id)
        ).one_or_none()
        if not collection:
            raise errors.NotFoundError("Collection not found")

        try:
            old_domain = collection.domain
            old_wildcard = bool(old_domain and old_domain.wildcard)
        except AttributeError:
            raise errors.NotFoundError("No domain is configured for this collection.")

        domain_name = allocate_domain(c.CollectionType(collection.collection_type))

        def delete_domain(collection_model):
            collection_model.delete_domain(new_domain=domain_name)

        update_object(collection, delete_domain)
        g.updated_objects.append(str(collection.id))

        if old_wildcard:
            _update_guest_collections_without_custom_domains(collection)

    res = api_schema.result_schema.dump(Result(message="Updated collection"))

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


@blueprint.route("/collections/<uuid:collection_id>/domain", methods=["GET"])
@authenticated
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
def collection_domain_get(collection_id: uuid.UUID) -> Response:
    """
    Get the custom domain document associated with this collection.

    This requires an `administrator` role on the Endpoint
    """
    with current_app.appsync_client.pause_notifications():
        collection = m.Collection.query.filter(
            m.Collection.id == str(collection_id)
        ).one_or_none()
        if not collection:
            raise errors.NotFoundError("Collection not found")
        old_domain = getattr(collection, "domain", None)

        if not old_domain:
            raise errors.NotFoundError("No domain is configured for this collection.")

    res = api_schema.result_schema.dump(
        Result(data=[api_schema.domain_schema.dump(old_domain)])
    )

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


def _update_guest_collections_without_custom_domains(collection: m.Collection) -> None:

    for guest in m.Collection.query.filter(
        ~m.Collection.deleted,  # type: ignore
        m.Collection.is_guest_collection,
        m.Collection.mapped_collection_id == collection.id,
    ):
        if not guest.has_domain:
            update_object(guest, m.Collection.update_domain_name, guest.domain_name)
            g.updated_objects.append(guest.id)
