from flask import (
    Blueprint,
    current_app,
    g,
    jsonify,
    request,
    Response,
)
import globus_sdk

from globus.manager.api.authentication.decorators import authenticated
from globus.manager.api.authorization import (
    RoleAuthorizer,
    requires_endpoint_visibility,
    requires_one_role,
)
from globus.manager.api.blueprints.decorators import expects_json
from globus.manager.api.blueprints.utils import (
    create_object,
    patch_object,
    put_object,
    update_object,
    validate_unique_custom_domain,
)
import globus.manager.api.errors as errors
import globus.manager.api.models as models
from globus.manager.api.models.result import Result
from globus.manager.api.schemas.api import (
    domain_schema,
    endpoint_owner_schema,
    endpoint_subscription_schema,
    endpoint_patch_schema,
    endpoint_schema,
    result_schema,
)
from globus.manager.api.utils.log import log_transfer_exception
from globus.manager.constants import PrincipalType, RoleType


blueprint = Blueprint("endpoint", __name__, url_prefix="/api/")


@blueprint.route("/endpoint", methods=["GET"])
@authenticated
@requires_endpoint_visibility
def endpoint_get() -> Response:
    """
    Get the Globus Connect Server's Endpoint document.
    """
    endpoint = models.Endpoint.query.one()

    res = result_schema.dump(Result(data=[endpoint_schema.dump(endpoint)]))

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


@blueprint.route("/endpoint", methods=["PATCH"])
@authenticated
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
@expects_json(schema=endpoint_patch_schema)
def endpoint_patch() -> Response:
    """
    Update the Endpoint document, changing only the properties included in the
    input. Items explicitly set to null in the input are removed from the
    Endpoint document. This operation optionally returns the Endpoint after
    applying the changes in the input if the include=endpoint query parameter
    is passed to this operation.
    """
    includes = request.args.get("include", "").split(",")
    include_endpoint = "endpoint" in includes

    current_subscription = current_app.subscription

    with current_app.appsync_client.pause_notifications():
        endpoint = models.Endpoint.query.one()

        input_endpoint_id = g.validated_data.get("endpoint_id")

        if input_endpoint_id and str(input_endpoint_id) != endpoint.endpoint_id:
            raise errors.ValidationError("Invalid id")

        new_subscription_id = g.validated_data.get("subscription_id")

        if "subscription_id" in g.validated_data:
            if new_subscription_id != current_subscription:
                _change_subscription(endpoint, new_subscription_id)

        g.validated_data.pop("subscription_id", None)
        patch_object(endpoint, g.validated_data)

        if not current_subscription and new_subscription_id:
            _add_admin_role_for_endpoint_owner(endpoint)
            _add_admin_role_for_collection_owner(endpoint)
            _update_collections()
        elif current_subscription and not new_subscription_id:
            _update_collections()

        if include_endpoint:
            result = [endpoint_schema.dump(endpoint)]
        else:
            result = []

        res = result_schema.dump(
            Result(
                data=result, message="Updated endpoint {}".format(endpoint.endpoint_id)
            )
        )

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    g.updated_objects.append(str(endpoint.endpoint_id))

    return response


@blueprint.route("/endpoint", methods=["PUT"])
@authenticated
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
@expects_json(schema=endpoint_schema)
def endpoint_put() -> Response:
    """
    Update the Endpoint document, replacing all properties with those in the
    input. This operation optionally returns the Endpoint after the update if
    the include=endpoint query parameter is passed to this operation.
    """
    includes = request.args.get("include", "").split(",")
    include_endpoint = "endpoint" in includes

    current_subscription = current_app.subscription

    with current_app.appsync_client.pause_notifications():
        endpoint = models.Endpoint.query.one()

        input_endpoint_id = g.validated_data.get("endpoint_id")
        if input_endpoint_id and str(input_endpoint_id) != endpoint.endpoint_id:
            raise errors.ValidationError("Invalid id")

        new_subscription_id = g.validated_data.get("subscription_id")

        if "subscription_id" in g.validated_data:
            if new_subscription_id != current_subscription:
                _change_subscription(endpoint, new_subscription_id)

        put_object(endpoint, g.validated_data)

        if not current_subscription and new_subscription_id:
            _add_admin_role_for_endpoint_owner(endpoint)
            _add_admin_role_for_collection_owner(endpoint)
            _update_collections()
        elif current_subscription and not new_subscription_id:
            _update_collections()

    if include_endpoint:
        result = [endpoint_schema.dump(endpoint)]
    else:
        result = []
    res = result_schema.dump(
        Result(message="Updated endpoint {}".format(endpoint.endpoint_id), data=result)
    )

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    g.updated_objects.append(str(endpoint.endpoint_id))

    return response


@blueprint.route("/endpoint/owner", methods=["PUT"])
@authenticated
@requires_one_role(role_names=["endpoint:owner"])
@expects_json(schema=endpoint_owner_schema)
def endpoint_owner_put() -> Response:
    """
    Assign a new identity to act as the endpoint owner.
    """
    new_owner_id = str(g.validated_data["identity_id"])

    new_owner_id_result = current_app.auth_client.get_identities(
        ids=[new_owner_id]
    ).data.get("identities", [])

    if len(new_owner_id_result) != 1:
        current_app.logger.error("Unable to resolve identity_id=%s", new_owner_id)
        raise errors.BadRequestError(
            "Unable to resolve identity_id to a Globus Auth identity"
        )

    new_owner_id = new_owner_id_result[0]["id"]
    new_owner_username = new_owner_id_result[0]["username"]

    if new_owner_username.endswith("@clients.auth.globus.org"):
        current_app.logger.error(
            "identity_id resolves to non-user %s", new_owner_username
        )
        raise errors.BadRequestError("The identity_id must belong to a user")

    with current_app.appsync_client.pause_notifications():
        owner_role = models.Role.query.filter(
            ~models.Role.deleted, models.Role.role == RoleType.endpoint_owner  # type: ignore
        ).one()
        old_owner_id = owner_role.identity_id
        endpoint = models.Endpoint.query.one()

        if new_owner_id != owner_role.identity_id:
            update_object(owner_role, lambda x, y: x.put_owner_role(y), new_owner_id)
            old_admin_role = models.Role.query.filter(
                ~models.Role.deleted,  # type: ignore
                models.Role.principal_type == PrincipalType.identity,
                models.Role.identity_id == old_owner_id,
                models.Role.role == RoleType.endpoint_administrator,
            ).one_or_none()

            if old_admin_role and old_admin_role.primary_role:

                def clear_primary(role):
                    role.primary_role = False

                update_object(old_admin_role, clear_primary)
                g.updated_objects.append(old_admin_role.id)

            if current_app.subscription:
                _add_admin_role_for_endpoint_owner(endpoint)
                _add_admin_role_for_collection_owner(endpoint)

        res = result_schema.dump(
            Result(
                message="Updated endpoint owner to {} ({})".format(
                    new_owner_id, new_owner_username
                )
            )
        )
    response = jsonify(res)
    response.status_code = res["http_response_code"]
    g.updated_objects.append(str(owner_role.id))

    return response


@blueprint.route("/endpoint/subscription_id", methods=["PUT"])
@authenticated
@expects_json(schema=endpoint_subscription_schema)
def endpoint_subscription_id_put() -> Response:
    """
    Change the subscription_id of this endpoint. Because subscription
    is enforcement is handled in a separate service than GCS and an
    organization's subscription manager may not be the administrator
    of the endpoint, this API has allows for both role-based
    authorization and subscription manager based authorization.

    The authorization allows the following:

    Caller has a role but is not subscription manager
        Remove an existing subscription from an endpoint, even if the
        caller is not a manager for that subscription.

    Caller does not have a role but is a subscription manager
        Set the subscription_id to a subscription they manage on a
        currently-unmanaged endpoint or remove the subscription_id from
        the endpoint if it is one that they managed.

    Caller has a role and is a subscription manager
        Set the subscription_id to a subcription they manage on an endpoint
        even if it is currently managed by a subscription that the caller is
        not a manager of.
    """
    with current_app.appsync_client.pause_notifications():
        endpoint = models.Endpoint.query.one()
        old_subscription = getattr(endpoint, "subscription_id", None)
        new_subscription = g.validated_data["subscription_id"]

        _change_subscription(endpoint, new_subscription)

    if old_subscription is None and new_subscription is not None:
        _add_admin_role_for_endpoint_owner(endpoint)
        _add_admin_role_for_collection_owner(endpoint)
        _update_collections()
    elif old_subscription is not None and new_subscription is None:
        _update_collections()

    # Assume assistant will update vhost files when subscription state changes
    result = result_schema.dump(
        Result(message="Updated Endpoint {}".format(endpoint.id))
    )

    response = jsonify(result)
    response.status_code = result["http_response_code"]
    g.updated_objects.append(str(endpoint.id))

    return response


@blueprint.route("/endpoint/domain", methods=["PUT"])
@authenticated
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
@expects_json(schema=domain_schema)
def endpoint_domain_put():
    """
    Register a new custom domain and certificate to to be used to serve
    the endpoint or collection.

    The domain is used for the endpoint itself. If the `wildcard`
    property is set to true, then all all collections associated
    with that endpoint that do not have a custom domain will be
    updated to use subdomains of that domain. Otherwise, only the
    endpoint will use that domain.
    """
    with current_app.appsync_client.pause_notifications():
        endpoint = models.Endpoint.query.one()

        old_domain = getattr(endpoint, "domain", None)
        old_domain_name = old_domain.domain_name if old_domain else None
        old_wildcard = bool(old_domain and old_domain.wildcard)

        if g.validated_data["domain_name"] != endpoint.domain_name:
            validate_unique_custom_domain(g.validated_data["domain_name"])

        update_object(endpoint, models.Endpoint.put_domain, g.validated_data)
        g.updated_objects.append(str(endpoint.id))

        new_domain = endpoint.domain
        new_wildcard = new_domain.wildcard

        if (old_wildcard or new_wildcard) and old_domain_name != new_domain.domain_name:
            _update_collections_using_endpoint_subdomains(endpoint)

    res = result_schema.dump(
        Result(
            message="Updated endpoint to use domain {}".format(new_domain.domain_name)
        )
    )
    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


@blueprint.route("/endpoint/domain", methods=["DELETE"])
@authenticated
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
def endpoint_domain_delete():
    """
    Delete the custom endpoint domain. This will cause the endpoint
    to revert to using a data.globus.org domain for the GCS Manager
    and any collections which do not have custom domains associated
    with them.

    This requires an `administrator` role on the Endpoint.
    """
    with current_app.appsync_client.pause_notifications():
        endpoint = models.Endpoint.query.one()
        old_domain = getattr(endpoint, "domain", None)

        if not old_domain:
            raise errors.NotFoundError("No domain is configured for this endpoint.")

        update_object(endpoint, models.Endpoint.delete_domain)
        g.updated_objects.append(str(endpoint.id))

        if old_domain and old_domain.wildcard:
            _update_collections_using_endpoint_subdomains(endpoint)

        result = result_schema.dump(
            Result(message="Updated Endpoint {}".format(endpoint.id))
        )

    response = jsonify(result)
    response.status_code = result["http_response_code"]
    g.updated_objects.append(str(endpoint.id))

    return response


@blueprint.route("/endpoint/domain", methods=["GET"])
@authenticated
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
def endpoint_domain_get():
    """
    Get the custom domain document associated with this endpoint.

    This requires an `administrator` role on the Endpoint.
    """
    with current_app.appsync_client.pause_notifications():
        endpoint = models.Endpoint.query.one()
        domain = getattr(endpoint, "domain", None)

        if not domain:
            raise errors.NotFoundError("No domain is configured for this endpoint.")

    res = result_schema.dump(Result(data=[domain_schema.dump(domain)]))

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


def _change_subscription(endpoint: models.Endpoint, new_subscription: str) -> None:
    old_subscription = endpoint.subscription_id

    transfer_token = g.auth_state.dependent_token("transfer.api.globus.org")
    if not transfer_token:
        current_app.logger.debug("No dependent token for transfer is available")
        raise errors.InvalidCredentialError(
            error_description=(
                "Unable to authenticate with the globus tranfer service on "
                "your behalf"
            )
        )
    authorizer = globus_sdk.AccessTokenAuthorizer(transfer_token)

    my_transfer_client = globus_sdk.TransferClient(
        authorizer=authorizer, environment=current_app.config["ENVIRONMENT"]
    )

    try:
        res = my_transfer_client.get("/private/my_subscriptions")
    except globus_sdk.TransferAPIError as e:
        log_transfer_exception(e)
        raise errors.BadGatewayError(
            "Error accessing subscription information: {}".format(e.code)
        )

    my_subscriptions = [s["id"] for s in res.data["DATA"]]

    if new_subscription == "DEFAULT":
        if len(my_subscriptions) == 1:
            new_subscription = my_subscriptions[0]
        elif len(my_subscriptions) > 1:
            raise errors.BadRequestError(
                "Unable to use DEFAULT subscription. "
                "Use one of {}".format(my_subscriptions)
            )
        else:
            raise errors.BadRequestError(
                "Unable to use DEFAULT subscription. "
                "Your identity does not manage any subscriptions"
            )

    if new_subscription is not None and new_subscription not in my_subscriptions:
        raise errors.BadRequestError(
            "Your identity does not manage subscription {}".format(new_subscription)
        )

    role_authorizer = RoleAuthorizer(["endpoint:administrator", "endpoint:owner"])
    role_identities = role_authorizer.authorize_identities(g.auth_state)

    current_app.logger.debug(
        "old_subscription=%s new_subscription=%s id_with_role=%s",
        old_subscription,
        new_subscription,
        role_identities,
    )
    if old_subscription is not None:
        if not (role_identities or old_subscription in my_subscriptions):
            raise errors.ForbiddenError(
                "You must have an endpoint owner or administrator role, or "
                "be a subscription manager to remove the subscription "
                "from this endpoint"
            )

    if new_subscription is None:
        if not (role_identities or old_subscription in my_subscriptions):
            raise errors.ForbiddenError(
                "You must have an endpoint owner or administrator role, or "
                "be a subscription manager to remove a subscription "
                "from this endpoint"
            )
    elif new_subscription is not None:
        if not (role_identities or new_subscription in my_subscriptions):
            raise errors.ForbiddenError(
                "You must have an endpoint owner or administrator role, or "
                "be a subscription manager to add a subscription "
                "to this endpoint"
            )

    def update_subscription(endpoint, new_subscription):
        resource = "/endpoint/{}/subscription".format(endpoint.endpoint_id)
        old_sub = endpoint.subscription_id

        if old_sub is not None and old_sub not in my_subscriptions:
            current_app.logger.debug("Attempting to clear subscription as client")
            try:
                subscription_doc = {"subscription_id": None}
                current_app.transfer_client.update_endpoint(
                    endpoint.endpoint_id, subscription_doc
                )
                endpoint.subscription_id = None
                current_app.logger.debug("Clear attempt is successful")
            except globus_sdk.TransferAPIError as e:
                log_transfer_exception(e)
                raise errors.BadRequestError("Unable to set subscription")

        try:
            current_app.logger.debug("Attempting to set new subscription as user")
            subscription_doc = {"subscription_id": new_subscription}
            my_transfer_client.put(resource, json_body=subscription_doc)
            endpoint.subscription_id = new_subscription
            current_app.logger.debug("Set subscription successful")
        except globus_sdk.TransferAPIError as e:
            log_transfer_exception(e)

    update_object(endpoint, update_subscription, new_subscription)
    current_app.subscription = new_subscription

    if endpoint.subscription_id != new_subscription:
        raise errors.BadRequestError("Unable to set subscription")


def _add_admin_role_for_endpoint_owner(endpoint: models.Endpoint) -> None:
    """
    Create an endpoint:administrator role on this endpoint for the
    identity which currently has an endpoint:owner role
    """
    owner_role = models.Role.query.filter(
        models.Role.role == RoleType.endpoint_owner
    ).one()

    admin_role = models.Role.query.filter(
        models.Role.role == RoleType.endpoint_administrator,
        models.Role.principal_type == PrincipalType.identity,
        models.Role.identity_id == owner_role.identity_id,
        models.Role.resource.is_(None),  # type: ignore
    ).one_or_none()

    if not admin_role:
        admin_role = models.Role(
            doc=dict(
                DATA_TYPE="role#1.0.0",
                principal=owner_role.principal,
                role="administrator",
            ),
            primary_role=True,
            endpoint=endpoint,
        )
        create_object(admin_role)
        g.created_objects.append(str(admin_role.id))


def _add_admin_role_for_collection_owner(endpoint: models.Endpoint):
    """
    Create an collection:administrator role on each mapped collection where the
    identity_id associated with the collection does not have an collection
    administrator role
    """
    for collection in models.Collection.query.filter(
        ~models.Collection.deleted, models.Collection.is_mapped_collection  # type: ignore
    ):
        admin_role = models.Role.query.filter(
            ~models.Role.deleted,  # type: ignore
            models.Role.role == RoleType.collection_administrator,
            models.Role.principal_type == PrincipalType.identity,
            models.Role.identity_id == collection.identity_id,
            models.Role.resource == collection.id,
        ).one_or_none()
        if not admin_role:
            admin_role = models.Role(
                doc=dict(
                    DATA_TYPE="role#1.0.0",
                    principal="urn:globus:auth:identity:" + collection.identity_id,
                    role="administrator",
                    collection=collection.id,
                ),
                primary_role=True,
                endpoint=endpoint,
            )
            create_object(admin_role)
            g.created_objects.append(str(admin_role.id))


def _update_collections() -> None:
    """
    Subscription state changed, so update all non deleted collection endpoint
    documents in transfer so that the https_url property is removed
    """
    for collection in models.Collection.query.filter(
        ~models.Collection.deleted  # type: ignore
    ):
        collection.update_transfer_endpoint()
        g.updated_objects.append(str(collection.id))


def _update_collections_using_endpoint_subdomains(endpoint: models.Endpoint) -> None:
    """
    Endpoint domain changed, so update all non deleted collections that do not
    have a custom domain.
    """
    for collection in models.Collection.query.filter(
        ~models.Collection.deleted,  # type:ignore
    ):
        if not (
            collection.has_domain
            or (
                collection.is_guest_collection
                and collection.mapped_collection.has_wildcard_domain
            )
        ):
            update_object(
                collection, models.Collection.update_domain_name, collection.domain_name
            )
            g.updated_objects.append(collection.id)
