from typing import Dict, List, Optional
import uuid

from flask import (
    Blueprint,
    Response,
    current_app,
    g,
    jsonify,
    request,
    url_for,
)
from werkzeug.exceptions import HTTPException

from globus.manager.api.authentication import authenticated
from globus.manager.api.blueprints.decorators import (
    Base62Marker,
    expects_json,
    paginated,
    query_parameter,
)
from globus.manager.api.blueprints.utils import update_object
from globus.manager.api.models.result import Result
from globus.manager.api.authorization import requires_one_role
import globus.manager.api.errors as errors
import globus.manager.api.models as models
import globus.manager.api.schemas.api as api_schema


blueprint = Blueprint("sharing_policies", __name__, url_prefix="/api/")


def get_collection_id_from_query() -> Optional[models.Collection]:
    collection_id = request.args.get("collection_id")
    collection = None
    collection = models.Collection.query.filter(
        ~models.Collection.deleted,  # type: ignore
        models.Collection.id == str(collection_id),
    ).one_or_none()

    g.collection = collection
    return collection


def get_collection_id_from_policy_doc() -> Optional[models.Collection]:
    try:
        json = request.get_json()
    except HTTPException:
        json = None

    if not json:
        raise errors.BadRequestError("Expected json input body.")

    collection_id = json.get("collection_id")
    collection = None

    if collection_id:
        collection = models.Collection.query.filter(
            ~models.Collection.deleted,  # type: ignore
            models.Collection.id == str(collection_id),
        ).one_or_none()

    g.collection = collection
    return collection


def get_collection_id_from_sharing_policy() -> Optional[models.Collection]:
    sharing_policy = None
    collection = None
    try:
        sharing_policy_id = str(request.view_args["sharing_policy_id"])
        sharing_policy = models.SharingPolicy.query.filter(
            models.SharingPolicy.id == sharing_policy_id
        ).one()
        collection = sharing_policy.collection
    except Exception as e:
        current_app.logger.debug("Couldn't find policy %s", str(e))
        raise errors.NotFoundError()

    if sharing_policy:
        g.collection = collection
        g.sharing_policy = sharing_policy

    return collection


@blueprint.route("/sharing_policies", methods=["GET"])
@authenticated
@requires_one_role(
    role_names=[
        "endpoint:owner",
        "endpoint:administrator",
        "collection:administrator:{collection_id}",
    ],
    collection_selector=get_collection_id_from_query,
)
@query_parameter(name_spec="username")
@paginated
def sharing_policies_get(username: Optional[str] = None) -> Response:
    collection = g.collection
    if not collection or collection.is_guest_collection:
        raise errors.NotFoundError()

    marker = g.marker
    page_size = g.page_size
    next_page_token = None

    query = models.SharingPolicy.query.filter(
        models.SharingPolicy.collection_id == collection.id
    )

    if marker:
        query = query.filter(models.ConfigurationObject.id >= marker.marker)

    query = query.order_by(models.ConfigurationObject.id)

    results = []  # type: List[Dict]
    for sp in query:
        if sp.users and username and username not in sp.users:
            continue

        if len(results) < page_size:
            results.append(api_schema.sharing_policy_schema.dump(sp))
        else:
            next_page_token = Base62Marker(marker_data=sp.id)
            break

    res = api_schema.result_schema.dump(Result(data=results, marker=next_page_token))

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


@blueprint.route("/sharing_policies", methods=["POST"])
@authenticated
@requires_one_role(
    role_names=[
        "endpoint:owner",
        "endpoint:administrator",
        "collection:administrator:{collection_id}",
    ],
    collection_selector=get_collection_id_from_policy_doc,
)
@expects_json(schema=api_schema.sharing_policy_schema)
def sharing_policies_post() -> Response:
    collection = g.collection
    if not collection or collection.is_guest_collection:
        raise errors.NotFoundError()
    policy_doc = g.validated_data  # type: api_schema.SharingPolicyLoadDict

    assert policy_doc is not None

    if "id" in policy_doc:
        policy_doc.pop("id")

    if not policy_doc.get("users"):
        raise errors.ValidationError("Missing users value in input document")

    sharing_policy = None  # type: Optional[models.SharingPolicy]

    def update_collection(c):
        nonlocal sharing_policy
        sharing_policy = collection.create_sharing_policy(policy_doc)

    update_object(collection, update_collection)
    assert sharing_policy is not None

    g.updated_objects.append(str(collection.id))
    g.created_objects.append(str(sharing_policy.id))

    res = api_schema.result_schema.dump(
        Result(
            http_response_code=201,
            message="Created sharing policy {}".format(sharing_policy.id),
            data=[api_schema.sharing_policy_schema.dump(sharing_policy)],
        )
    )

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    response.headers["Location"] = url_for(
        ".sharing_policies_get_by_id",
        sharing_policy_id=sharing_policy.id,
        _external=True,
        _scheme="https",
    )

    return response


@blueprint.route("/sharing_policies/<uuid:sharing_policy_id>", methods=["GET"])
@authenticated
@requires_one_role(
    role_names=[
        "endpoint:owner",
        "endpoint:administrator",
        "collection:administrator:{collection_id}",
    ],
    collection_selector=get_collection_id_from_sharing_policy,
)
def sharing_policies_get_by_id(sharing_policy_id: uuid.UUID) -> Response:
    collection = g.collection

    assert collection is not None and collection.is_mapped_collection

    sharing_policy = g.sharing_policy

    assert sharing_policy is not None

    res = api_schema.result_schema.dump(
        Result(data=[api_schema.sharing_policy_schema.dump(sharing_policy)])
    )

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


@blueprint.route("/sharing_policies/<uuid:sharing_policy_id>", methods=["DELETE"])
@authenticated
@requires_one_role(
    role_names=[
        "endpoint:owner",
        "endpoint:administrator",
        "collection:administrator:{collection_id}",
    ],
    collection_selector=get_collection_id_from_sharing_policy,
)
def sharing_policies_delete_by_id(sharing_policy_id: uuid.UUID) -> Response:
    assert g.sharing_policy is not None
    assert g.collection is not None and g.collection.is_mapped_collection

    sharing_policy_id = g.sharing_policy.id

    if sharing_policy_id == models.SharingPolicy.default_collection_policy(
        g.collection.id
    ):
        raise errors.BadRequestError(
            "The default sharing policy for a collection can not be deleted"
        )

    def delete_policy(collection):
        collection.delete_sharing_policy(sharing_policy_id)

    update_object(g.collection, delete_policy)

    res = api_schema.result_schema.dump(
        Result(message="Deleted sharing policy {}".format(str(sharing_policy_id)))
    )
    g.deleted_objects.append(sharing_policy_id)

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response
