from flask import (
    Blueprint,
    current_app,
    g,
    jsonify,
    request,
    url_for,
)

from globus.manager.api.authentication.decorators import authenticated
from globus.manager.api.blueprints.decorators import (
    Base62Marker,
    expects_json,
    paginated,
)
from globus.manager.api.blueprints.utils import (
    check_subscription,
    create_object,
    delete_object,
    patch_object,
    put_object,
    update_object,
)
import globus.manager.api.errors as errors
from globus.manager.api.models.result import Result

import globus.manager.api.models as models
from globus.manager.api.models import db
from globus.manager.api.schemas.api.account import account_schema
from globus.manager.api.schemas.api.result import result_schema
from globus.manager.api.schemas.api.storage_gateway import (
    storage_gateway_schema,
    storage_gateway_patch_schema,
    storage_gateway_private_schema,
)
from globus.manager.api.authorization import (
    requires_one_role,
    AuthenticationTimeoutFilter,
    IdentitySet,
    PublicAuthorizer,
    RoleAuthorizer,
    RoleSpec,
    StorageGatewayAuthorizer,
)


blueprint = Blueprint("storage_gateways", __name__, url_prefix="/api/")


class StorageGatewayPaginationMarker(Base62Marker):
    def __init__(
        self, storage_gateway_id=None, identity_id=None, username=None, marker=None
    ):
        self._storage_gateway_id = storage_gateway_id
        self._identity_id = identity_id or None
        self._username = username or None
        marker_data = self.__marker_data()

        super().__init__(marker=marker, marker_data=marker_data)

    @property
    def marker(self):
        return self.__marker_data()

    @marker.setter
    def marker(self, value):
        storage_gateway_id, identity_id, username = value.split(":", 2)
        self._storage_gateway_id = storage_gateway_id
        self._identity_id = identity_id or None
        self._username = username or None

    @property
    def storage_gateway_id(self):
        return self._storage_gateway_id or ""

    @storage_gateway_id.setter
    def storage_gateway_id(self, value):
        self._storage_gateway_id = value

    @property
    def identity_id(self):
        return self._identity_id or ""

    @identity_id.setter
    def identity_id(self, value):
        self._identity_id = value or None

    @property
    def username(self):
        return self._username or ""

    @username.setter
    def username(self, value):
        self._username = value or None

    def __marker_data(self):
        if not self._storage_gateway_id:
            return None

        return "{}:{}:{}".format(
            self._storage_gateway_id, self._identity_id or "", self._username or ""
        )


def _verify_role_used_mfa(role_names):
    """
    Check if any of g.authorized_identities used MFA in this session.

    This assumes a previously successful call to @requires_one_role() which
    populated g.auth_state without the MFA requirement. The caller has decided that
    MFA is required. If any of the authorized_identities (which are assumed to
    also have the necessary role) used MFA, then return success.

    If none of the roles have used MFA, then we'll get the list of identities
    that have any of the necessary roles, regardless of their current session
    characteristics and error with an authentication timeout requiring HA+MFA.
    """
    for i in g.authorized_identities:
        if g.auth_state.multi_factor_authentications.get(i["id"]):
            return

    role_specs = []
    for role_name in role_names:
        role_specs.append(RoleSpec(role_name))
    role_authorizer = RoleAuthorizer(role_specs)

    authorized_identities = role_authorizer.authorize_identities(g.auth_state)

    raise errors.AuthenticationTimeoutError(
        identities=authorized_identities,
        high_assurance=True,
        require_mfa=True,
        error_description=(
            "You must authenticate using multi-factor authentication in order to "
            "enable MFA requirements on a storage gateway."
        ),
    )


@blueprint.route("/storage_gateways", methods=["GET"])
@blueprint.route("/storage_gateways/<uuid:storage_gateway_id>", methods=["GET"])
@paginated(marker_class=StorageGatewayPaginationMarker)
@authenticated
def storage_gateways_get(storage_gateway_id=None):
    """
    Get the Endpoint's list of StorageGateway documents.

    The include query parameter controls what additional information is
    included in the Result document.

    This operation requires either the endpoint to have the public property set
    to true, the caller to have a role that allows viewing this Endpoint, the
    user to have an identity which is allowed by the individual Storage Gateway
    policies, or an identity which has an ACL for a collection created on this
    Storage Gateway.

    Authorization

    RoleAuthorizer:

    - endpoint:owner
    - endpoint:administrator
    - endpoint:activity_manager
    - endpoint:activity_monitor
    - collection:administrator:{collection_id}
    - collection:activity_manager:{collection_id}
    - collection:activity_monitor:{collection_id}
    - collection:access_manager:{collection_id}
    - CollectionACLAuthorizer:

        - {collection_id}

    - StorageGatewayAuthorizer

        - {storage_gateway_id}

    Query Parameters:

    include: List of document types to include in the response.

    If private_policies is in the list, then include private storage gateway
    policies in the attached storage_gateways document. This requires an
    administrator or own role on the Endpoint, and the caller must have
    authenticated that identity according to the policies of the storage Gateway.

    If accounts is included in the list, then the caller's list of local
    accounts is returned for each storage gateway is included in the data.
    """
    includes = request.args.get("include", "").split(",")
    include_private = "private_policies" in includes
    include_accounts = "accounts" in includes
    page_size = g.page_size
    marker = g.marker

    endpoint_is_public = PublicAuthorizer().authorize_identities(g.auth_state)

    endpoint_admin_authorizer = RoleAuthorizer(
        ["endpoint:owner", "endpoint:administrator"]
    )
    caller_endpoint_admin_identities = endpoint_admin_authorizer.authorize_identities(
        g.auth_state
    )
    caller_is_endpoint_admin = bool(caller_endpoint_admin_identities)

    endpoint_role_authorizer = RoleAuthorizer(
        ["endpoint:activity_manager", "endpoint:activity_monitor"]
    )
    caller_has_endpoint_role = endpoint_role_authorizer.authorize_identities(
        g.auth_state
    )

    collection_role_authorizer = RoleAuthorizer(
        [
            "collection:administrator:*",
            "collection:activity_manager:*",
            "collection:activity_monitor:*",
            "collection:access_manager:*",
        ]
    )
    caller_has_collection_role = collection_role_authorizer.authorize_identities(
        g.auth_state
    )

    # TODO: ACL authorizer check
    caller_has_any_acl = False

    endpoint_is_visible = bool(
        endpoint_is_public
        or caller_is_endpoint_admin
        or caller_has_endpoint_role
        or caller_has_collection_role
        or caller_has_any_acl
    )

    current_app.logger.info(
        (
            "Authorization results: endpoint_public:{} endpoint_admin:{} "
            "endpoint_role:{} collection_role:{} acl:{}"
        ).format(
            endpoint_is_public,
            caller_is_endpoint_admin,
            caller_has_endpoint_role,
            caller_has_collection_role,
            caller_has_any_acl,
        )
    )

    if not caller_is_endpoint_admin:
        include_private = False

    new_marker = None
    result_data = []
    sgquery = models.StorageGateway.query.filter(~models.StorageGateway.deleted)

    if storage_gateway_id:
        storage_gateway_id = str(storage_gateway_id)
        sgquery = sgquery.filter_by(id=storage_gateway_id)
    elif marker:
        current_app.logger.debug("selection > %s", marker)
        sgquery = sgquery.filter(
            models.ConfigurationObject.id >= marker.storage_gateway_id
        )

    sgquery = sgquery.order_by(models.StorageGateway.id)

    result_size = 0
    for sg in sgquery:
        sg_visible = (
            endpoint_is_public or caller_is_endpoint_admin or caller_has_endpoint_role
        )

        if not sg_visible:
            if storage_gateway_id:
                assert sg.id == storage_gateway_id
            # Check for a identity allowed by the storage gateway policies
            sg_authorizer = StorageGatewayAuthorizer(sg)
            sg_authorized_identities = sg_authorizer.authorize_identities(g.auth_state)
            if bool(sg_authorized_identities):
                sg_visible = True

        if not sg_visible:
            current_app.logger.debug("Searching for collections on SG={}".format(sg.id))
            for col in models.Collection.query.filter(
                ~models.Collection.deleted,
                models.Collection.storage_gateway_id == sg.id,
            ):
                # Check for a collection role for this storage gateway
                current_app.logger.debug("Found {}, checking roles".format(col.id))
                sg_role_authorizer = RoleAuthorizer(
                    [
                        RoleSpec("collection:administrator:{collection_id}", col),
                        RoleSpec("collection:activity_manager:{collection_id}", col),
                        RoleSpec("collection:activity_monitor:{collection_id}", col),
                        RoleSpec("collection:access_manager:{collection_id}", col),
                    ]
                )
                collection_authorized_identities = sg_role_authorizer.authorize_identities(
                    g.auth_state
                )
                current_app.logger.debug(
                    "authorized_identities={}".format(collection_authorized_identities)
                )
                if bool(collection_authorized_identities):
                    sg_visible = True
                    break

        if not sg_visible:
            continue

        if include_private:
            # If caller is an endpoint admin or owner, they may be able to
            # see private data, but only if it is within the HA requirements of
            # the storage gateway being inspected
            sg_filter = AuthenticationTimeoutFilter(
                require_session=sg.high_assurance,
                authentication_timeout=sg.authentication_timeout_mins,
                require_mfa=sg.require_mfa,
            )

            sg_identity_set = IdentitySet(caller_endpoint_admin_identities)
            if sg_filter(sg_identity_set, g.auth_state):
                sg_schema = storage_gateway_private_schema
            else:
                sg_schema = storage_gateway_schema
        else:
            sg_schema = storage_gateway_schema

        if (
            marker is None
            or marker.storage_gateway_id != sg.id
            or not marker.identity_id
        ):
            if result_size == page_size:
                new_marker = StorageGatewayPaginationMarker(storage_gateway_id=sg.id)
                break

            sg_result = sg_schema.dump(sg)

            result_data.append(sg_result)
            result_size += 1

        if include_accounts:
            # Filter identity set based on storage policies
            authorizer = StorageGatewayAuthorizer(
                sg, enforce_authentication_timeout=False
            )
            authorized_identities = authorizer.authorize_identities(g.auth_state)

            # Map the authorized identity set to allowed local usernames if any
            accounts = []
            try:
                accounts = sg.map_identities(authorized_identities)
            except errors.InvalidUserError as e:
                detail = e.detail

                if detail:
                    current_app.logger.info(
                        "Identity mapped to unusable username(s) %s",
                        getattr(detail, "usernames", []),
                    )

            accounts.sort(key=lambda a: a.identity_id + a.username)

            if marker is not None and marker.storage_gateway_id == sg.id:
                accounts = [
                    account
                    for account in accounts
                    if (
                        account.identity_id > marker.identity_id
                        or (
                            account.identity_id == marker.identity_id
                            and account.username >= marker.username
                        )
                    )
                ]
            for local_account in accounts:
                if result_size == page_size:
                    new_marker = StorageGatewayPaginationMarker(
                        storage_gateway_id=sg.id,
                        identity_id=local_account.identity_id,
                        username=local_account.username,
                    )
                    break

                account = account_schema.dump(local_account)
                result_data.append(account)
                result_size += 1

    if (not endpoint_is_visible) and len(result_data) == 0:
        raise errors.ForbiddenError("Permission denied")

    if storage_gateway_id is not None and len(result_data) == 0:
        raise errors.NotFoundError()

    res = result_schema.dump(
        Result(data=result_data, marker=new_marker.encode() if new_marker else None)
    )

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


@blueprint.route("/storage_gateways", methods=["POST"])
@authenticated
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
@expects_json(schema=storage_gateway_schema)
def storage_gateways_post():
    """
    POST /storage_gateways

    Create a StorageGateway on the Endpoint. On success, this operation returns
    a copy of the created Storage Gateway with the system generated id added.

    Authorization
    RoleAuthorizer:
        endpoint:owner
        endpoint:administrator
    """
    sg_model_class = models.StorageGateway.polymorphic_class(
        g.validated_data["connector_id"]
    )

    # May be other reasons sg_model_class is none?
    if not sg_model_class:
        raise errors.ValidationError(
            'The connector "{}" is not known to the endpoint. Either your client '
            "is misconfigured or the service requires an update to support it.".format(
                str(g.validated_data["connector_id"])
            )
        )

    check_subscription(
        sg_model_class.connector,
        high_assurance=g.validated_data.get("high_assurance", False),
    )

    if g.validated_data.get("require_mfa"):
        _verify_role_used_mfa(["endpoint:owner", "endpoint:administrator"])

    for non_settable in ["id", "process_user", "load_dsi_module", "deleted"]:
        g.validated_data.pop(non_settable, None)

    storage_gateway_id = None
    existing_with_display_name = []
    with current_app.appsync_client.pause_notifications():
        for sg in models.StorageGateway.query.filter(
            ~models.StorageGateway.deleted,
            models.StorageGateway.display_name == g.validated_data["display_name"],
        ):
            existing_with_display_name.append(sg.id)
        if existing_with_display_name:
            raise errors.ConflictError(
                "One or more storage gateways with this display_name already " "exist.",
                resource_ids=existing_with_display_name,
            )
        try:
            sg = sg_model_class(doc=g.validated_data)
            db.session.add(sg)
            db.session.commit()

            current_app.logger.debug(
                "Created storage gateway model {}: {}".format(sg, sg.configuration_hash)
            )
            storage_gateway_id = sg.id
        except (TypeError, ValueError) as e:
            current_app.logger.error(
                "Error creating storage gateway model: {}".format(str(e))
            )
            raise errors.ValidationError(error_description=str(e))

        create_object(sg)

        result = storage_gateway_schema.dump(sg)

        res = result_schema.dump(
            Result(
                http_response_code=201,
                message="Created new Storage Gateway {}".format(sg.id),
                data=[result],
            )
        )

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    response.headers["Location"] = url_for(
        ".storage_gateways_get",
        storage_gateway_id=storage_gateway_id,
        _external=True,
        _scheme="https",
    )
    g.created_objects.append(storage_gateway_id)

    return response


@blueprint.route("/storage_gateways/<uuid:storage_gateway_id>", methods=["DELETE"])
@authenticated
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
def storage_gateways_delete(storage_gateway_id):
    sgid = str(storage_gateway_id)

    with current_app.appsync_client.pause_notifications():
        sg = models.StorageGateway.query.filter(
            models.StorageGateway.id == sgid, ~models.StorageGateway.deleted
        ).one_or_none()

        if not sg:
            raise errors.NotFoundError()

        existing_user_credentials = []
        existing_deleted_user_credentials = []

        for u in models.UserCredential.query.filter(
            models.UserCredential.storage_gateway_id == sg.id
        ).all():
            if u.deleted:
                existing_deleted_user_credentials.append(u.id)
            else:
                existing_user_credentials.append(u.id)

        existing_collections = []
        existing_deleted_collections = []

        for c in models.Collection.query.filter(
            models.Collection.storage_gateway_id == sg.id
        ):
            if c.deleted:
                existing_deleted_collections.append(c.id)
            else:
                existing_collections.append(c.id)

        if existing_collections:
            message = (
                "StorageGateway {} ({}) is in use by "
                "{} UserCredential(s) and {} Collection(s)".format(
                    sg.display_name,
                    sg.id,
                    len(existing_user_credentials),
                    len(existing_collections),
                )
            )
            current_app.logger.error("DELETE: {}".format(message))
            detail = "inuse"
            raise errors.ResourceInUseError(message, detail=detail)

        while existing_user_credentials:
            u = existing_user_credentials.pop()
            try:
                uc = models.UserCredential.query.filter(
                    models.UserCredential.id == u
                ).one_or_none()
                if uc:
                    delete_object(uc, update=True)
                    g.deleted_objects.append(u)
            except Exception as e:
                current_app.logger.debug("Error deleting user_credential %s: %s", u, e)
                message = (
                    "StorageGateway {} ({}) is in use by "
                    "{} UserCredential(s) and {} Collection(s)".format(
                        sg.display_name,
                        sg.id,
                        len(existing_user_credentials) + 1,
                        len(existing_collections),
                    )
                )
                raise errors.ResourceInUseError(message)

        delete_object(sg, update=True)

    res = result_schema.dump(Result(message="Deleted Storage Gateway {}".format(sgid)))

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    g.deleted_objects.append(sgid)

    return response


def get_storage_gateway_from_path():
    storage_gateway_id = request.view_args.get("storage_gateway_id")
    if not storage_gateway_id:
        return None
    storage_gateway = models.StorageGateway.query.filter(
        ~models.StorageGateway.deleted,
        models.StorageGateway.id == str(storage_gateway_id),
    ).one_or_none()

    g.storage_gateway = storage_gateway
    return storage_gateway


@blueprint.route("/storage_gateways/<uuid:storage_gateway_id>", methods=["PATCH"])
@authenticated
@requires_one_role(
    role_names=["endpoint:owner", "endpoint:administrator"],
    storage_gateway_selector=get_storage_gateway_from_path,
)
@expects_json(schema=storage_gateway_patch_schema)
def storage_gateways_patch(storage_gateway_id):
    """
    Update a Storage Gateway, change only the properties included in the input
    document. It returns a document containing the Storage Gateway after the
    changes have been applied. Items explicitly set to null in the input are
    removed from the Storage Gateway document.

    Some properties are immutable, in general, the `id`,  `connector_id`, and
    `high_assurance` properties cannot be changed, though storage
    gateways may enforce additional restrictions.

    This operation may return a Conflict error if any collections exist which
    would be not be consistent with the change in Storage Gateway policies.
    """
    for non_settable in ["process_user", "load_dsi_module", "deleted"]:
        g.validated_data.pop(non_settable, None)

    sg_doc = g.validated_data
    storage_gateway_id = str(storage_gateway_id)

    with current_app.appsync_client.pause_notifications():
        sg = g.storage_gateway
        assert sg is not None

        old_auth_timeout = sg.authentication_timeout_mins
        old_domains = sg.allowed_domains
        old_require_mfa = sg.require_mfa

        if (
            sg.high_assurance
            and not old_require_mfa
            and g.validated_data.get("require_mfa")
        ):
            _verify_role_used_mfa(["endpoint:owner", "endpoint:administrator"])

        if "id" in sg_doc and str(sg_doc["id"]) != str(storage_gateway_id):
            raise errors.ValidationError("Invalid id")

        patch_object(sg, sg_doc)

        res_doc = storage_gateway_schema.dump(
            models.StorageGateway.query.filter_by(id=storage_gateway_id).one()
        )
        res = result_schema.dump(
            Result(
                message="Updated Storage Gateway {}".format(storage_gateway_id),
                data=[res_doc],
            )
        )

        changed_timeout = old_auth_timeout != sg.authentication_timeout_mins
        changed_domains = old_domains != sg.allowed_domains
        changed_mfa = old_require_mfa != sg.require_mfa
        if changed_timeout or changed_domains or changed_mfa:
            _update_collections(
                storage_gateway_id, changed_timeout, changed_domains, changed_mfa
            )

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    g.updated_objects.append(str(storage_gateway_id))

    return response


@blueprint.route("/storage_gateways/<uuid:storage_gateway_id>", methods=["PUT"])
@authenticated
@requires_one_role(
    role_names=["endpoint:owner", "endpoint:administrator"],
    storage_gateway_selector=get_storage_gateway_from_path,
)
@expects_json(schema=storage_gateway_schema)
def storage_gateways_put(storage_gateway_id):
    for non_settable in ["process_user", "load_dsi_module", "deleted"]:
        g.validated_data.pop(non_settable, None)

    sg_doc = g.validated_data
    storage_gateway_id = str(storage_gateway_id)

    with current_app.appsync_client.pause_notifications():
        sg = g.storage_gateway
        assert sg is not None

        old_auth_timeout = sg.authentication_timeout_mins
        old_domains = sg.allowed_domains
        old_require_mfa = sg.require_mfa

        if "id" not in sg_doc or str(sg_doc["id"]) != str(storage_gateway_id):
            raise errors.ValidationError("Invalid id")

        if (
            sg.high_assurance
            and not old_require_mfa
            and g.validated_data.get("require_mfa")
        ):
            _verify_role_used_mfa(["endpoint:owner", "endpoint:administrator"])

        put_object(sg, sg_doc)

        res_doc = storage_gateway_schema.dump(sg)
        res = result_schema.dump(
            Result(
                message="Updated Storage Gateway {}".format(storage_gateway_id),
                data=[res_doc],
            )
        )

        changed_timeout = old_auth_timeout != sg.authentication_timeout_mins
        changed_domains = old_domains != sg.allowed_domains
        changed_mfa = old_require_mfa != sg.require_mfa
        if changed_timeout or changed_domains or changed_mfa:
            _update_collections(
                storage_gateway_id, changed_timeout, changed_domains, changed_mfa
            )

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    g.updated_objects.append(str(storage_gateway_id))

    return response


def _update_collections(
    storage_gateway_id: str,
    changed_timeout: bool,
    changed_domains: bool,
    changed_mfa: bool,
) -> None:
    """
    Check for collections related to a particular Storage Gateway's ID, and
    update them in transfer and appsync. This is called when a storage gateway
    update changes a piece of configuration that must be reflected in the
    related collections.

    :param str storage_gateway_id:
        The ID of the Storage Gateway that has been updated. Every collection
        created on this Storage Gateway will be updated.
    """
    updated_objects = []
    for collection in models.Collection.query.filter(
        ~models.Collection.deleted,  # type: ignore
        models.Collection.storage_gateway_id == storage_gateway_id,
    ):

        def update_collection(collection):
            if changed_timeout or changed_mfa:
                collection.update_transfer_endpoint()
            if changed_domains:
                collection.update_scopes()

        # Note: The data_access scope is relevant only to mapped collections
        if (
            changed_timeout
            or (changed_domains and collection.is_mapped_collection)
            or changed_mfa
        ):
            update_object(collection, update_collection)
            updated_objects.append(collection.id)
    g.updated_objects.extend(updated_objects)
