from flask import current_app, Blueprint, g, jsonify, request, url_for
from sqlalchemy.sql.expression import false

from globus.manager.api.blueprints.decorators import (
    Base62Marker,
    expects_json,
    paginated,
)
import globus.manager.api.errors as errors

from globus.manager.api.authentication.decorators import authenticated
from globus.manager.api.authorization import (
    requires_credential_owner,
    requires_storage_gateway_policy,
    IdentitySet,
    PublicAuthorizer,
    StorageGatewayAuthorizer,
    EndpointRoleAuthorizer,
)
import globus.manager.api.models as models
from globus.manager.api.models import db, Result
from globus.manager.api.blueprints.decorators import (
    server_timing,
    db_read,
)
from globus.manager.api.schemas.api.result import result_schema
from globus.manager.api.schemas.api.user_credential import (
    user_credential_patch_schema,
    user_credential_schema,
)
from globus.manager.api.blueprints.utils import (
    create_object,
    delete_object,
    patch_object,
    put_object,
)


blueprint = Blueprint("user_credentials", __name__, url_prefix="/api")


@blueprint.route("/user_credentials", methods=["GET"])
@paginated
@server_timing
@authenticated
def user_credentials_get():
    """
    Get User Credentials

    The caller may only retrieve User Credentials which were
    created by an identity in the caller's identity set unless they are an
    administrator. If the endpoint is not public, returns ForbiddenError
    if the user has no credentials

    Authorizations:
    -   PublicAuthorizer
    -   StorageGatewayAuthorizer (if "storage_gateway" given)
    """
    storage_gateway = request.args.get("storage_gateway")
    marker = g.marker
    page_size = g.page_size

    endpoint_is_public = PublicAuthorizer().authorize_identities(g.auth_state)

    if storage_gateway:
        sg = models.StorageGateway.query.filter(
            models.StorageGateway.id == storage_gateway
        ).one_or_none()
        if not sg:
            raise errors.NotFoundError()
        authorizer = StorageGatewayAuthorizer(sg, enforce_authentication_timeout=False)
        authorized_identities = authorizer.authorize_identities(g.auth_state)
    else:
        authorized_identities = IdentitySet()
        for i in g.auth_state.identity_set_data:
            authorized_identities.append(i)

    # List of all credentials which are not deleted, owned by the caller, and
    # which are not locally instantiated dummy credentials
    query = models.UserCredential.query.filter(
        models.UserCredential.identity_id.in_([i["id"] for i in authorized_identities]),
        models.UserCredential.deleted == false(),
    )

    if storage_gateway:
        query = query.filter_by(storage_gateway_id=storage_gateway)

    if marker:
        query = query.filter(models.ConfigurationObject.id >= marker.marker)

    query = query.order_by(models.ConfigurationObject.id).limit(page_size + 1)

    g.server_timing.start(db_read)
    query_results = []
    for credential in query:
        for i in authorized_identities:
            if i["id"] == credential.identity_id:
                if credential.storage_gateway.map_identity(i) == credential.username:
                    current_app.logger.debug(
                        "Adding credential %s to response", credential
                    )
                    query_results.append(credential)
                    break
                else:
                    current_app.logger.debug("Ignoring credential without mapping")
                    break
    g.server_timing.finish(db_read)

    # If the caller has an endpoint role, they can see all collections
    caller_has_endpoint_role = EndpointRoleAuthorizer.authorize_identities(g.auth_state)

    if query_results == [] and not endpoint_is_public and not caller_has_endpoint_role:
        raise errors.ForbiddenError()

    results = []
    for uc in query_results:
        results.append(user_credential_schema.dump(uc))

    if len(results) > page_size:
        next_page_token = Base62Marker(marker_data=results.pop()["id"])
    else:
        next_page_token = None

    res = result_schema.dump(Result(data=results, marker=next_page_token))

    resp = jsonify(res)
    resp.status_code = res["http_response_code"]
    resp.headers["Server-Timing"] = g.server_timing.as_string()

    return resp


def get_user_credential_id_from_path():
    """
    Resolves the user_credential_id view parameter to a UserCredential if
    a UserCredential with that ID is in the database.
    """
    user_credential_id = request.view_args.get("user_credential_id")

    assert user_credential_id is not None

    user_credential = None
    if user_credential_id:
        user_credential = models.UserCredential.query.filter(
            models.UserCredential.id == str(user_credential_id)
        ).one_or_none()

    g.user_credential = user_credential
    return user_credential


@blueprint.route("/user_credentials/<uuid:user_credential_id>", methods=["GET"])
@authenticated
@requires_credential_owner(credential_selector=get_user_credential_id_from_path)
def user_credentials_get_by_id(user_credential_id):
    """
    Get a User Credential by its id.

    The caller must have the identity_id of the user credential
    in its identity set.

    Authorizations:
    -   CredentialOwnerAuthorizer
    """
    uc = g.user_credential

    result = user_credential_schema.dump(uc)

    res = result_schema.dump(Result(data=[result]))

    resp = jsonify(res)
    resp.status_code = res["http_response_code"]

    return resp


def get_storage_gateway_from_credential_doc():
    storage_gateway_id = g.validated_data.get("storage_gateway_id")
    if storage_gateway_id:
        storage_gateway = models.StorageGateway.query.filter(
            models.StorageGateway.id == str(storage_gateway_id)
        ).one_or_none()
        g.storage_gateway = storage_gateway
        return storage_gateway


@blueprint.route("/user_credentials/<uuid:user_credential_id>", methods=["PATCH"])
@authenticated
@expects_json(schema=user_credential_patch_schema)
@requires_credential_owner(credential_selector=get_user_credential_id_from_path)
def user_credentials_patch(user_credential_id):
    """
    Update a user credential on a Storage Gateway. This is required
    for some connectors that require a local user name or other
    credential information to access the storage system. See the
    connector-specific documentation for details on what is needed.

    The caller must have the identity_id of the user credential
    in its identity set.

    Authorizations:
    -   CredentialOwnerAuthorizer
    """
    uc = g.user_credential

    with current_app.appsync_client.pause_notifications():
        if uc.deleted:
            current_app.logger.error("PATCH %s: already deleted", str(uc.id))
            raise errors.NotFoundError()
        elif g.validated_data.get("deleted"):
            raise errors.ValidationError("Cannot change deleted via PATCH")

        patch_object(uc, g.validated_data)
        res = result_schema.dump(
            Result(
                message="Updated User Credential {}".format(str(uc.id)),
                data=[user_credential_schema.dump(uc)],
            )
        )

    resp = jsonify(res)
    resp.status_code = res["http_response_code"]
    g.updated_objects.append(uc.id)

    return resp


@blueprint.route("/user_credentials/<uuid:user_credential_id>", methods=["PUT"])
@authenticated
@expects_json(schema=user_credential_schema)
@requires_credential_owner(credential_selector=get_user_credential_id_from_path)
def user_credentials_put(user_credential_id):
    """
    Update a user credential on a Storage Gateway. This is required
    for some connectors that require a local user name or other
    credential information to access the storage system. See the
    connector-specific documentation for details on what is needed.

    The caller must have the identity_id of the user credential
    in its identity set.

    Authorizations:
    -   CredentialOwnerAuthorizer
    """
    uc = g.user_credential

    with current_app.appsync_client.pause_notifications():
        if uc.deleted:
            current_app.logger.error("PUT %s: already deleted", str(uc.id))
            raise errors.NotFoundError()
        elif g.validated_data.get("deleted", False) != g.user_credential.deleted:
            raise errors.ValidationError("Cannot change deleted via PUT")

        put_object(uc, g.validated_data)
        res = result_schema.dump(
            Result(
                message="Updated User Credential {}".format(str(uc.id)),
                data=[user_credential_schema.dump(uc)],
            )
        )

    resp = jsonify(res)
    resp.status_code = res["http_response_code"]
    g.updated_objects.append(uc.id)

    return resp


@blueprint.route("/user_credentials", methods=["POST"])
@authenticated
@expects_json(schema=user_credential_schema)
@requires_storage_gateway_policy(
    storage_gateway_selector=get_storage_gateway_from_credential_doc
)
def user_credentials_post():
    """
    Create a user credential on a Storage Gateway. This is required
    for some connectors that require a local user name or other
    credential information to access the storage system. See the
    connector-specific documentation for details on what is needed.

    The caller is authorized based on the StorageGateway identity
    policies, so users with no assigned roles on the endpoint may
    be permitted to access this operation.

    Authorizations:
    -   StorageGatewayAuthorizer
    """
    assert g.validated_data is not None
    assert g.storage_gateway is not None

    authorized_identity_ids = sorted([i["id"] for i in g.authorized_identities])

    if not authorized_identity_ids:
        raise errors.InvalidUserError()

    if "identity_id" not in g.validated_data:
        if len(authorized_identity_ids) == 1:
            g.validated_data["identity_id"] = authorized_identity_ids[0]
        else:
            raise errors.ValidationError(
                "Missing identity_id property, and you have more than one "
                "authorized Globus account."
            )

    requested_identity_id = str(g.validated_data["identity_id"])

    if requested_identity_id not in authorized_identity_ids:
        raise errors.IdNotInIdentitySetError(id=requested_identity_id)

    authorized_identity = [
        i for i in g.authorized_identities if i["id"] == requested_identity_id
    ][0]

    mapped_to = g.storage_gateway.map_identity(authorized_identity)

    if not mapped_to:
        raise errors.InvalidUserError()

    requested_username = g.validated_data.get("username")
    if requested_username is None:
        g.validated_data["username"] = mapped_to

    if requested_username and requested_username != mapped_to:
        raise errors.InvalidUserError(usernames=[mapped_to])
    if not g.storage_gateway.check_user_allowed(mapped_to):
        raise errors.InvalidUserError(usernames=[mapped_to])

    with current_app.appsync_client.pause_notifications():
        try:
            cls = models.UserCredential.polymorphic_class(
                g.storage_gateway.connector_id
            )
            old_uc = cls.query.filter(
                cls.storage_gateway_id == g.storage_gateway.id,
                cls.username == g.validated_data["username"],
                cls.identity_id == requested_identity_id,
            ).one_or_none()

            if old_uc:
                if not old_uc.deleted:
                    raise errors.ResourceExistsError()
            elif not old_uc:
                uc = cls(doc=g.validated_data, storage_gateway=g.storage_gateway)
        except ValueError as e:
            current_app.logger.error("Error creating using credential: {}".format(e))
            raise errors.ValidationError()

        if old_uc:
            old_uc.deleted = False
            put_object(old_uc, g.validated_data)
            uc = old_uc
        else:
            create_object(uc)

        credential_json = user_credential_schema.dump(uc)

        res = result_schema.dump(
            Result(
                http_response_code=201,
                message="Created User Credential {}".format(uc.id),
                data=[credential_json],
            )
        )

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    response.headers["Location"] = url_for(
        ".user_credentials_get_by_id",
        user_credential_id=uc.id,
        _external=True,
        _scheme="https",
    )
    g.created_objects.append(uc.id)

    return response


@blueprint.route("/user_credentials/<uuid:user_credential_id>", methods=["DELETE"])
@authenticated
@requires_credential_owner(credential_selector=get_user_credential_id_from_path)
def user_credentials_delete(user_credential_id):
    """
    Delete a User Credential by its id.

    The caller must have the identity_id of the user credential in its identity
    set.

    Authorizations:
    -   CredentialOwnerAuthorizer
    """
    uc = g.user_credential

    with current_app.appsync_client.pause_notifications():
        if uc.deleted:
            current_app.logger.error("DELETE %s: already deleted", str(uc.id))
            raise errors.NotFoundError()

        existing_collections = [
            uc.id
            for c in db.session.query(models.Collection.id)
            .filter(models.Collection.user_credential_id == uc.id)
            .all()
        ]

        if existing_collections:
            current_app.logger.info(
                "DELETE %s: in use by %d Collection(s)",
                uc.id,
                len(existing_collections),
            )

        delete_object(uc, update=True)

        res = result_schema.dump(
            Result(
                message="Deleted User Credential {}".format(str(user_credential_id)),
                data=[],
            )
        )

    resp = jsonify(res)
    resp.status_code = res["http_response_code"]
    g.deleted_objects.append(uc.id)

    return resp
