import os
import secrets
import time
from typing import Any, Dict, List, Mapping, Optional, Union
import uuid

from flask import current_app, render_template
from jwcrypto.jwk import JWK
import rsa
import yaml

import globus_sdk
from globus_sdk.config import get_service_url
from globus_sdk.exc import GlobusAPIError

import globus.manager.api.errors as errors
from globus.manager.api.schemas.api.domain import DomainDict
from globus.manager.api.schemas.appsync.oidc_server import (
    OIDCServerDict,
    oidc_server_yaml_private_schema,
)
from globus.manager.api.utils.log import log_auth_exception
from globus.manager.utils.run_as_user import run_as_user

from .domain import Domain


class OIDCClient:
    """
    This class represents an OAuth client which can interact with the
    OIDCServer.
    """

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        redirect_uris: List[List[Optional[str]]],
    ):
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uris = redirect_uris

    def __repr__(self):
        return "<OIDCClient {}>".format(self.client_id)


class OIDCServer:
    """
    This class is the model for managing the optional OIDC server configuration
    for an endpoint. It includes methods for interacting with Globus Auth to
    register this service as well for writing the service configuration to the
    filesystem.
    """

    _DATA_TYPES = ["oidc_server#1.0.0"]
    _DATA_TYPE = _DATA_TYPES[0]
    OIDC_SERVER_PROPERTIES = {
        "oidc_server#1.0.0": [
            "client_id",
            "client_secret",
            "domain_name",
            "environment",
            "display_name",
            "support_contact",
            "support_email",
            "pam_service",
            "shared_key",
            "shared_session_key",
            "clients",
            "domain",
            "jwk",
            "deleted",
        ]
    }

    def __init__(
        self,
        appsync_data: Optional[OIDCServerDict] = None,
        doc: Optional[OIDCServerDict] = None,
    ):
        """
        The constructor can be called with either the appsync_data or
        the doc options set. In either case, the value should be an
        OIDCServerDict containing the configuration parameters. When
        passed as the appsync_data option, this class assumes that
        all data has already been validated and treats this more like
        an object restore from the database. When passed as the doc
        option, this method validates the document contents, and creates
        resources as needed to fill in the configuration.
        """
        self.client_id = None  # type: Optional[str]
        self.client_secret = None  # type: Optional[str]
        self._domain_name = None  # type: Optional[str]
        self.environment = None  # type: Optional[str]
        self.display_name = None  # type: Optional[str]
        self.support_contact = None  # type: Optional[str]
        self.support_email = None  # type: Optional[str]
        self.pam_service = None  # type: Optional[str]
        self.shared_key = None  # type: Optional[str]
        self.shared_session_key = None  # type: Optional[str]
        self._clients = {}  # type: Dict[str, OIDCClient]
        self._domain = None  # type: Optional[Domain]
        self._jwk = None  # type: Optional[JWK]
        self.deleted = False  # type: bool

        self._auth_client_data = {}  # type: Mapping[str, Any]
        self._auth_identity_provider_data = {}  # type: Mapping[str, Any]

        if appsync_data is not None:
            if "domain" in appsync_data:
                domain = Domain(appsync_data=appsync_data["domain"])
                self._put(
                    DATA_TYPE=appsync_data["DATA_TYPE"],
                    **{
                        k: v
                        for k, v in appsync_data.items()
                        if k not in ["DATA_TYPE", "domain"]
                    },
                    domain=domain
                )
            else:
                self._put(**appsync_data)
        elif doc is not None:
            self.put(**doc)

            if self.deleted:
                return

            if self.environment is None:
                self.environment = current_app.environment

            if self.shared_key is None:
                self.shared_key = secrets.token_urlsafe(32)

            if self.shared_session_key is None:
                self.shared_session_key = secrets.token_urlsafe(32)

            if self._jwk is None:
                self._jwk = JWK.generate(kty="RSA", size=1024)

            if self.client_id is None:
                self._create_child_client()

            if self.client_secret is None:
                self._create_client_secret()

            if self.domain_name not in self.auth_client_data.get("fqdns", []):
                self._register_fqdn()

            if not self.clients:
                # globus-oidc won't respond to the autodiscovery endpoint
                # unless there is at least one client
                self.add_client()
        else:
            raise errors.InternalServerError("Invalid constructor call")

    @property
    def domain(self) -> Domain:
        """
        The domain property contains a Domain object that has information
        about the domain name, certificate chain, and private key to
        configure the virtual host for this object.

        The setter for this property can be passed in data in either
        the JSON representation or an object. In the former case, the
        value is subject to validation by the Domain constructor.
        """
        if self._domain is None:
            raise AttributeError(
                "The domain property is not defined for this OIDCServer"
            )
        else:
            return self._domain

    @domain.setter
    def domain(self, value: Union[DomainDict, Domain]) -> None:
        if isinstance(value, Domain):
            self._domain = value
        else:
            self._domain = Domain(doc=value)

    @property
    def domain_name(self) -> str:
        """
        The domain_name property is a read-only string property derived from
        the domain property. It contains the fully-qualified domain name that
        the globus-oidc service is using.
        """
        if self._domain is not None:
            return self._domain.domain_name
        else:
            raise AttributeError(
                "The domain_name property is not defined for this OIDCServer"
            )

    @property
    def clients(self) -> Mapping[str, OIDCClient]:
        """
        The clients property contains the list of OAuth clients (and their
        secrets) which can access this OIDC service.
        """
        return self._clients

    @clients.setter
    def clients(
        self, value: Optional[Mapping[str, Union[OIDCClient, Mapping[str, Any]]]]
    ) -> None:
        clients = {}
        if value is not None:
            for k, v in value.items():
                if isinstance(v, OIDCClient):
                    clients[k] = v
                else:
                    clients[k] = OIDCClient(k, v["client_secret"], v["redirect_uris"])
        self._clients = clients

    @property
    def jwk(self) -> Dict[str, Any]:
        """
        The jwk property contains an RSA keypair used by the globus-oidc service in
        JWK encoding.
        """
        if self._jwk is not None:
            return self._jwk.export(private_key=True, as_dict=True)
        else:
            raise AttributeError("The jwk property for this OIDCServer is not defined")

    @jwk.setter
    def jwk(self, value: Dict[str, Any]):
        self._jwk = JWK(**value)

    @property
    def auth_client_data(self) -> Mapping[str, Any]:
        """
        The auth_client_data property contains a mapping with client document from
        auth for the client_id associated with this OIDCServer object.
        """
        if not self.client_id:
            raise AttributeError(
                "The auth_client_data property is not defined for this OIDCServer"
            )

        if not self._auth_client_data:
            try:
                auth_response = current_app.auth_client.get(
                    "/v2/api/clients/{}".format(self.client_id)
                )
                current_app.logger.debug(
                    "Get client from auth returned {0}".format(
                        auth_response.http_status
                    )
                )
                if not auth_response.data.get("client", {}).get("id"):
                    raise errors.BadGatewayError(
                        "Error retrieving Globus Auth client. Please check "
                        "that the credentials are correct."
                    )

                self._auth_client_data = auth_response.data["client"]
                current_app.logger.debug("Retrieved client with ID %s", self.client_id)
            except GlobusAPIError as e:
                msg = (
                    "Error retrieving Globus Auth client: {0} {1}. Please check "
                    "that the credentials are correct and Globus Auth is accessible."
                ).format(e.http_status, str(e))
                raise Exception(msg)
        return self._auth_client_data

    @property
    def auth_identity_provider_data(self) -> Mapping[str, Any]:
        """
        The auth_client_data property contains a mapping with identity_provider
        document from auth for the domain_name used by this OIDCServer object.
        """
        if not self._auth_identity_provider_data:
            try:
                resp = current_app.auth_client.get(
                    "/v2/api/identity_providers", params={"domains": [self.domain_name]}
                )
                data = resp.data["identity_providers"]

                if data:
                    self._auth_identity_provider_data = data[0]
                else:
                    raise AttributeError(
                        "The auth_identity_provider_data property is not "
                        "set on this OIDCServer"
                    )
            except GlobusAPIError as e:
                log_auth_exception(e)

                raise AttributeError(
                    "The auth_identity_provider_data property is not "
                    "set on this OIDCServer"
                )
        return self._auth_identity_provider_data

    @property
    def autodiscovery_url(self) -> str:
        """
        The autodiscovery_url is a read-only property containing the well-known
        OIDC configuration service for this service.
        """
        return "https://{}/.well-known/openid-configuration".format(self.domain_name)

    def add_client(self) -> OIDCClient:
        """
        The add_client method generates a new OAuth client id and secret to allow
        Globus Auth to interact with this service.

        :return:
            This method returns an OIDCClient object containing the information
            about the new client. It also adds the object to the values in the
            clients property.
        """
        client_id = str(uuid.uuid4())
        client_secret = secrets.token_urlsafe(32)
        redirect_uris = [
            [
                "{}p/authenticate/callback".format(
                    get_service_url(current_app.config["ENVIRONMENT"], "auth")
                ),
                None,
            ]
        ]
        client = self._clients[client_id] = OIDCClient(
            client_id, client_secret, redirect_uris
        )
        return client

    def _put(self, DATA_TYPE: str = None, **doc) -> None:
        """
        The _put method updates the contents of this object with a new
        document. This method does not check the document for validity, and
        ignores unknown properties.

        :param DATA_TYPE:
            The DATA_TYPE parameter is the versioned document data type that
            is being processed. If this is unknown, the object will use process
            this as the newest document type it understands.
        :param **doc:
            The doc parameter contains the rest of the configuration document.
        """
        properties = self.OIDC_SERVER_PROPERTIES.get(DATA_TYPE) if DATA_TYPE else None
        if properties is None:
            properties = self.OIDC_SERVER_PROPERTIES[self._DATA_TYPE]

        for prop in properties:
            if prop in doc:
                if prop == "domain_name":
                    # Ignore this and use the domain object's value
                    continue
                setattr(self, prop, doc[prop])

    def put(self, DATA_TYPE: str = None, **doc) -> None:
        """
        The put method updates the contents of this object with a new
        document. This method does checking of the document for validity and
        will only process documents of a known data type.

        :param DATA_TYPE:
            The DATA_TYPE parameter is the versioned document data type that
            is being processed. If this is unknown, the object will use process
            this as the newest document type it understands.
        :param **doc:
            The doc parameter contains the rest of the configuration document.
        """
        if DATA_TYPE not in self._DATA_TYPES:
            raise ValueError(
                "Unsupported DATA_TYPE for OIDCServer: {}".format(DATA_TYPE)
            )
        if "client_id" in doc and self.client_id and doc["client_id"] != self.client_id:
            raise errors.ValidationError("The client_id property is immutable")
        if (
            "environment" in doc
            and self.environment
            and doc["environment"] != self.environment
        ):
            raise errors.ValidationError("The environment property is immutable")
        if "domain" in doc and self._domain is not None:
            if doc["domain"]["domain_name"] != self.domain.domain_name:
                raise errors.ValidationError("The domain_name property is immutable")
        if "domain" in doc:
            doc["domain"] = Domain(doc=doc["domain"])
        self._put(DATA_TYPE, **doc)

    def _patch(self, DATA_TYPE: str = None, **doc) -> None:
        """
        The _patch method updates the contents of this object with a change
        document.  This method does not check the document for validity, and
        ignores unknown properties.

        :param DATA_TYPE:
            The DATA_TYPE parameter is the versioned document data type that
            is being processed. If this is unknown, the object will use process
            this as the newest document type it understands.
        :param **doc:
            The doc parameter contains the rest of the configuration document.
        """
        properties = self.OIDC_SERVER_PROPERTIES.get(DATA_TYPE) if DATA_TYPE else None
        if properties is None:
            properties = self.OIDC_SERVER_PROPERTIES[self._DATA_TYPE]

        for prop in properties:
            if prop in doc:
                if prop == "domain_name":
                    # Ignore this and use the domain object's value
                    continue
                elif prop == "domain":
                    self.domain.patch(**doc[prop])
                else:
                    setattr(self, prop, doc[prop])

    def patch(self, DATA_TYPE: str = None, **doc) -> None:
        """
        The patch method updates the contents of this object with a new
        document. This method does checking of the document for validity and
        will only process documents of a known data type. Any items missing
        in the input document will retain their previous values.

        :param DATA_TYPE:
            The DATA_TYPE parameter is the versioned document data type that
            is being processed. If this is unknown, the object will use process
            this as the newest document type it understands.
        :param **doc:
            The doc parameter contains the rest of the configuration document.
        """
        if DATA_TYPE not in self._DATA_TYPES:
            raise errors.ValidationError("Unsupported DATA_TYPE {}".format(DATA_TYPE))

        if "client_id" in doc and self.client_id and doc["client_id"] != self.client_id:
            raise errors.ValidationError("The client_id property is immutable")
        if (
            "environment" in doc
            and self.environment
            and doc["environment"] != self.environment
        ):
            raise errors.ValidationError("The environment property is immutable")
        if "domain" in doc and "domain_name" in doc["domain"]:
            if doc["domain"]["domain_name"] != self.domain.domain_name:
                raise errors.ValidationError("The domain_name property is immutable")

        if "display_name" in doc:
            request_body = {
                "client": {"name": doc["display_name"]}
            }  # type: Dict[str, Any]
            try:
                auth_response = current_app.auth_client.put(
                    "/v2/api/clients/{}".format(self.client_id), json_body=request_body
                )
                self._auth_client_data = auth_response["client"]
            except GlobusAPIError as e:
                log_auth_exception(e)
                raise errors.BadGatewayError(
                    "Unable to update display_name: {}".format(str(e))
                )

        self._patch(DATA_TYPE, **doc)

    def _create_child_client(self) -> None:
        """
        Create a new auth client for this service as a child of the Endpoint's
        auth client.
        """
        current_app.logger.debug("Creating Globus Auth identity....")

        request_body = {
            "client": {
                "name": self.display_name,
                "public_client": False,
                "redirect_uris": [],
                "visibility": "private",
                "required_idp": None,
                "preselect_idp": None,
            }
        }  # type: Dict[str, Any]
        try:
            auth_response = current_app.auth_client.post(
                "/v2/api/clients", json_body=request_body
            )
            current_app.logger.debug(
                "Created client with auth returned {0}".format(
                    auth_response.http_status
                )
            )
            assert auth_response.http_status == 201
            assert auth_response.data["client"] and auth_response.data["client"]["id"]
            self._auth_client_data = auth_response.data["client"]
            self.client_id = self._auth_client_data["id"]
            current_app.logger.debug("Created client with ID %s", self.client_id)
        except GlobusAPIError as e:
            msg = (
                "Error creating Globus Auth client: {0} {1}. Please check "
                + "that the credentials are correct and Globus Auth is accessible."
            ).format(e.http_status, str(e))
            raise errors.BadGatewayError(msg)

    def _create_client_secret(self) -> None:
        """
        Create a new client secret for the client_id associated with this service.
        """
        # Using the parent secret to create a child secret.
        if self.client_id is None:
            raise errors.InternalServerError(
                "The client_id property for this OIDCServer is not defined."
            )
        auth_child_client = globus_sdk.ConfidentialAppAuthClient(
            self.client_id,
            current_app.client_secret,
            environment=current_app.environment,
        )

        request_body = {"credential": {"name": self.display_name}}

        # Create the secret
        try:
            auth_response = auth_child_client.post(
                "/v2/api/clients/{}/credentials".format(self.client_id),
                json_body=request_body,
            )
            current_app.logger.debug(
                "Created auth credentials returned {0}".format(
                    auth_response.http_status
                )
            )
            if not auth_response.data.get("credential", {}).get("secret"):
                raise errors.BadGatewayError(
                    "Invalid response from Globus Auth creating a client secret"
                )
            self.client_secret = auth_response.data["credential"]["secret"]
        except GlobusAPIError as e:
            log_auth_exception(e)
            msg = (
                "Error creating Globus Auth credentials: {0} {1}. Please check "
                "that the credentials are correct and Globus Auth is accessible."
            ).format(e.http_status, str(e))
            raise errors.BadGatewayError(msg)

    def _register_fqdn(self) -> None:
        """
        Register the domain_name of this OIDC server with this service's
        Globus Auth client_id.
        """
        current_app.logger.debug("Registering FQDN with Auth")

        # FQDN isn't registered yet with Auth - register it
        RETRIES = 10
        for attempt in range(1, RETRIES):
            try:
                current_app.logger.debug(
                    "Attempting to register FQDN {} with Globus Auth client {} "
                    "using parent client ID {}...".format(
                        self.domain_name, self.client_id, current_app.client_id,
                    )
                )
                auth_response = current_app.auth_client.post(
                    "/v2/api/clients/{client_id}/fqdns".format(
                        client_id=self.client_id
                    ),
                    json_body={"fqdn": self.domain_name},
                )
                current_app.logger.debug(
                    "Registered dns with auth returned {0}".format(
                        auth_response.http_status
                    )
                )
                assert auth_response.http_status in [200, 201]
                break
            except GlobusAPIError as e:
                msg = (
                    "Error registering FQDN with Globus Auth: {} {}. Please check "
                    "that a DNS TXT record exists at {} with value '{}'."
                ).format(
                    e.http_status,
                    str(e),
                    self.domain_name,
                    current_app.auth_client.client_id,
                )
                current_app.logger.info(msg)

                # Auth should probably send a 400 in the domain already
                # registered case
                if e.http_status > 500:
                    raise errors.InternalServerError(msg)
                if attempt + 1 < RETRIES:
                    current_app.logger.debug(str(e))
                    time.sleep(20)
                else:
                    raise errors.InternalServerError(msg)

    def register_idp(self) -> str:
        """
        Register the IdP with Globus Auth. This creates a new identity provider
        for this OIDC Server's domain name. This requires the OIDC service to
        be running enough to serve the OIDC autodiscovery URL.
        """
        try:
            auth_identity_provider_data = self.auth_identity_provider_data

            if auth_identity_provider_data:
                return auth_identity_provider_data["id"]
        except AttributeError:
            pass

        if not self.clients:
            raise errors.InternalServerError("Misconfigured OIDC service")

        client = list(self.clients.values())[0]
        params = {
            "identity_provider": {
                "discovery_url": self.autodiscovery_url,
                "name": self.display_name,
                "client_credentials": {
                    "client_id": client.client_id,
                    "client_secret": client.client_secret,
                },
                "attribute_map": {"username": "sub", "immutable_id": "sub"},
                "contact_email": self.support_email,
                "contact_name": self.support_contact,
            }
        }

        auth_child_client = globus_sdk.ConfidentialAppAuthClient(
            self.client_id,
            current_app.client_secret,
            environment=current_app.environment,
        )
        try:
            resp = auth_child_client.post(
                "/v2/api/identity_providers", json_body=params
            )
            if resp.http_status in [200, 201]:
                current_app.logger.debug(resp)
                self._auth_identity_provider_data = resp["identity_provider"]
            else:
                raise errors.BadGatewayError(
                    "Unable to create IdP, Globus Auth returned status code {}.".format(
                        resp.http_status
                    )
                )
        except GlobusAPIError as e:
            log_auth_exception(e)
            raise errors.BadGatewayError(
                "Unable to create IdP, Globus Auth returned "
                "status code {}: {}.".format(e.http_status, str(e))
            )
        return resp["identity_provider"]["id"]

    @run_as_user(user="gcsweb")
    def _setup_vhost(self):
        """
        Configure a virtual host for this OIDC service. This writes the certificate
        and key data stored in the domain property as well as a vhost based on the
        service configuration template.
        """
        self.domain.write_config()

        vhost_conf_dir = os.path.join(current_app.config["HOME"], "etc/httpd/conf.d")
        vhost_conf_file = os.path.join(vhost_conf_dir, self.client_id + ".conf")

        vhost_conf_file_tmp = vhost_conf_file + ".gcs.tmp"

        if not os.path.exists(vhost_conf_dir):
            os.makedirs(vhost_conf_dir, 0o700)

        current_app.logger.debug("Write vhost to %s", vhost_conf_file_tmp)
        with open(
            os.open(vhost_conf_file_tmp, os.O_CREAT | os.O_WRONLY | os.O_TRUNC, 0o600),
            "w",
        ) as f:
            f.write(render_template("oidc.conf.jinja2", oidc_server=self))
        os.rename(vhost_conf_file_tmp, vhost_conf_file)

    @run_as_user(user="gcsweb")
    def _cleanup_vhost(self):
        """
        Remove a virtual host for this OIDC service. This deletes the
        certificate and key data stored in the domain property as well as the
        vhost configuration.
        """
        self.domain.write_config(delete=True)
        vhost_conf_file = os.path.join(
            current_app.config["HOME"], "etc/httpd/conf.d", self.client_id + ".conf"
        )
        if os.path.exists(vhost_conf_file):
            os.remove(vhost_conf_file)

    @run_as_user(user="gcsweb")
    def _setup_yaml(self):
        """
        Configure the config_appsync.yaml document for the OIDC server.
        """
        yamldir = os.path.expanduser("~globusoidc/globus-oidc")
        local_config = os.path.join(yamldir, "config_appsync.yaml")

        if os.path.exists(local_config):
            os.remove(local_config)

        with open(
            os.open(local_config, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, 0o0640), "w",
        ) as f:
            yaml.dump(oidc_server_yaml_private_schema.dump(self), f)

    @run_as_user(user="gcsweb")
    def _cleanup_yaml(self):
        """
        Remove the config_appsync.yaml document for the OIDC server.
        """
        yamldir = os.path.expanduser("~globusoidc/globus-oidc")
        local_config = os.path.join(yamldir, "config_appsync.yaml")

        if os.path.exists(local_config):
            os.remove(local_config)

    @run_as_user(user="gcsweb")
    def _setup_keys(self):
        """
        Write the RSA keypair stored in the jwk property to the filesystem for
        the OIDC service to use.
        """
        keydir = os.path.expanduser("~globusoidc/globus-oidc/keys")
        keyfile = os.path.join(keydir, "key.pem")
        keypubfile = keyfile + ".pub"

        jwk_json = self.jwk

        pubkey = rsa.PublicKey(
            self._jwk._decode_int(jwk_json["n"]), self._jwk._decode_int(jwk_json["e"])
        )

        privkey = rsa.PrivateKey(
            self._jwk._decode_int(jwk_json["n"]),
            self._jwk._decode_int(jwk_json["e"]),
            self._jwk._decode_int(jwk_json["d"]),
            self._jwk._decode_int(jwk_json["p"]),
            self._jwk._decode_int(jwk_json["q"]),
        )
        with open(
            os.open(keyfile, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, 0o0640), "wb"
        ) as f:
            f.write(privkey.save_pkcs1("PEM"))
        with open(
            os.open(keypubfile, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, 0o0640), "wb"
        ) as f:
            f.write(pubkey.save_pkcs1("PEM"))

    @run_as_user(user="gcsweb")
    def _cleanup_keys(self):
        """
        Remove the RSA keypair stored in the jwk property from the filesystem.
        """
        keydir = os.path.expanduser("~globusoidc/globus-oidc/keys")
        keyfile = os.path.join(keydir, "key.pem")
        keypubfile = keyfile + ".pub"

        if os.path.exists(keyfile):
            try:
                os.remove(keyfile)
            except Exception:
                pass

        if os.path.exists(keypubfile):
            try:
                os.remove(keypubfile)
            except Exception:
                pass

    def write_config(self, delete=False):
        """
        Write configuration files for the OIDC service so that it can run
        on the local node.
        """
        current_app.logger.info("Persisting OIDC configuration")
        if self.deleted:
            delete = True

        if not delete:
            current_app.logger.info("Writing keys")
            self._setup_keys()
            current_app.logger.info("Writing yaml")
            self._setup_yaml()
            current_app.logger.info("Writing vhost")
            self._setup_vhost()
        else:
            self._cleanup_keys()
            self._cleanup_yaml()
            self._cleanup_vhost()

    def delete(self):
        """
        Delete the Auth resources associated with this OIDC server.
        """
        if self.client_id:
            current_app.auth_client.delete("/v2/api/clients/{}".format(self.client_id))
            self._auth_client_data = {}
            self._auth_identity_provider_data = {}
        self.deleted = True

    def __repr__(self):
        return "<OIDCServer {}>".format(self.client_id)


__all__ = ["OIDCServer"]
