import re
from typing import Optional, Union
from typing_extensions import Literal

import marshmallow as ma
from mypy_extensions import TypedDict

from globus.manager.scripts.utils.dns_patterns import DOMAIN_NAME_PATTERN
from .data_type import DataTypeSchema

ManagedDomainDict = TypedDict(
    "ManagedDomainDict",
    {
        "DATA_TYPE": Literal["domain#1.0.0"],
        "domain_name": str,
        "wildcard": bool,
        "certificate": str,
        "certificate_chain": Optional[str],
        "private_key": str,
    },
    total=False,
)
UnmanagedDomainDict = TypedDict(
    "UnmanagedDomainDict",
    {
        "DATA_TYPE": Literal["domain#1.0.0"],
        "domain_name": str,
        "wildcard": bool,
        "certificate_path": str,
        "certificate_chain_path": Optional[str],
        "private_key_path": str,
    },
    total=False,
)
DomainDict = Union[ManagedDomainDict, UnmanagedDomainDict]


class DomainSchema(ma.Schema):
    _DATA_TYPE = "domain#1.0.0"
    DATA_TYPE = ma.fields.Str(
        default=_DATA_TYPE, validate=[ma.validate.Equal(_DATA_TYPE)]
    )
    id = ma.fields.UUID()
    domain_name = ma.fields.Str(
        required=True, validate=ma.validate.Regexp(DOMAIN_NAME_PATTERN)
    )
    wildcard = ma.fields.Boolean(required=True)
    certificate = ma.fields.Str(
        required=False,
        allow_none=True,
        validate=ma.validate.Regexp(
            "^(.*\n)?-----BEGIN CERTIFICATE-----.*", re.MULTILINE
        ),
    )
    certificate_chain = ma.fields.Str(
        required=False,
        allow_none=True,
        validate=ma.validate.Regexp(
            "^(.*\n)?-----BEGIN CERTIFICATE-----.*", re.MULTILINE
        ),
    )
    private_key = ma.fields.Str(
        required=False,
        allow_none=True,
        validate=ma.validate.Regexp(
            "^(.*\n)?-----BEGIN PRIVATE KEY-----.*", re.MULTILINE
        ),
    )
    certificate_path = ma.fields.Str(
        required=False, allow_none=True, validate=ma.validate.Regexp("^/.*")
    )
    certificate_chain_path = ma.fields.Str(
        required=False, allow_none=True, validate=ma.validate.Regexp("^/.*")
    )
    private_key_path = ma.fields.Str(
        required=False, allow_none=True, validate=ma.validate.Regexp("^/.*")
    )

    def dump(self, *args, **kwargs) -> DomainDict:
        return super().dump(*args, **kwargs)

    def load(self, *args, **kwargs) -> DomainDict:
        return super().load(*args, **kwargs)

    private = [
        "certificate_chain",
        "private_key",
        "certificate_path",
        "certificate_chain_path",
        "private_key_path",
    ]

    @ma.validates_schema
    def validate_managed_or_unmanaged(self, data, **kwargs):
        # A domain can be either managed or unmanaged.
        # For a managed domain:
        # - both the certificate and private_key must be present and be a valid
        #   certificate and key combination
        # - certificate_chain may be set to include a list of other certs
        # - both certificate_path and private_key_path must not be set or must be
        #   set to null.
        # For an unmanaged domain:
        # - both the certificate_path and private_key_path must be present and
        #   point to existing files.
        # - certificate_chain_path may optionally be set and if it is non-null
        #   must point to a valid certificate chain file.
        # - the certificate and private_key values must not be set or must be
        #   set to null.
        certificate = data.get("certificate")
        certificate_chain = data.get("certificate_chain")
        private_key = data.get("private_key")

        certificate_path = data.get("certificate_path")
        certificate_chain_path = data.get("certificate_chain_path")
        private_key_path = data.get("private_key_path")

        managed = certificate and private_key
        unmanaged = certificate_path and private_key_path

        if (
            (managed and unmanaged)
            or not (managed or unmanaged)
            or (managed and certificate_chain_path)
            or (unmanaged and certificate_chain)
        ):
            raise ma.ValidationError(
                "A domain#1.0.0 document must contain both certificate_chain and "
                "private_key or both certificate_path and private_key_path"
            )


domain_schema = DomainSchema(load_only=DomainSchema.private)
domain_private_schema = DomainSchema()

DataTypeSchema.register_schema(domain_schema)
