"""
globus-connect-server node cleanup
"""
import json
import os
import shutil
import textwrap

import click
from flask import current_app
from gcs_cli.parsing.command_decorator import globus_command

import globus_sdk

from globus_gcs.gridftp.gridftp_service import GridFTPService
from globus_gcs.service_manager import ServiceManager
from globus_gcs.apache import GCSManager, WebServerManager
from globus.manager.appsync.client import GraphQLError
from globus.manager.api.app import make_app
from globus_gcs.utils.host_info import public_ip
from globus.manager.utils.keychain import NodeKey
from globus.manager.appsync.client import ConfigurationObjectTag
from globus.manager.api.models import (
    db,
    LetsEncryptCertificate,
    Collection,
    Configuration,
    ConfigurationObject,
    Endpoint,
    GlobusDnsDomain,
    Role,
    StorageGateway,
    UserCredential,
)


@globus_command(
    "cleanup",
    short_help="Clean up a Globus Connect Server node",
    help="Clean up a Globus Connect Server node",
    disable_options=["format", "explicit_host"],
)
@click.option("--ip-address", "-i", help="IP Address of this GCS Node")
@click.option(
    "--local-cleanup-only",
    is_flag=True,
    help="Clean up local state only. Do not attempt to remove the node from "
    "the endpoint.",
)
@click.option(
    "--root",
    hidden=True,
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    default="/",
)
@click.option("--verbose", "-v", is_flag=True)
@click.pass_context
def node_cleanup(ctx, ip_address, local_cleanup_only, root, verbose):
    if os.getuid() != 0:
        click.echo("This command must be run as the root user", err=True)
        ctx.exit(1)

    if verbose:
        os.environ["GCS_MANAGER_LOG_LEVEL"] = "DEBUG"
    app = make_app(app_config="globus.manager.api.config.node_cli")
    db.init_app(app)

    with app.app_context():
        app.logger.info("node_cleanup: start")
        db.create_all()
        db.session.commit()

        cleanup = NodeCleanup(ip_address, local_cleanup_only, root)
        cleanup.run()
        app.logger.info("node_cleanup: end")


class NodeCleanup:
    def __init__(self, ip_address, local_cleanup_only, root):
        if ip_address is None:
            ip_address = public_ip()
        self.ip_address = ip_address
        self.environment = None
        self.client_id = None
        self.client_secret = None
        self.node_key = None

        environment = Configuration.query.filter_by(name="environment").one_or_none()
        if environment:
            self.environment = environment.value

        client_id = Configuration.query.filter_by(name="client_id").one_or_none()
        if client_id:
            self.client_id = client_id.value
        client_secret = Configuration.query.filter_by(
            name="client_secret"
        ).one_or_none()
        if client_secret:
            self.client_secret = client_secret.value
        node_key_data = Configuration.query.filter_by(name="node_key").one_or_none()
        if node_key_data:
            self.node_key = NodeKey(**json.loads(node_key_data.value))
        current_app.config["CLIENT_ID"] = self.client_id
        current_app.config["CLIENT_SECRET"] = self.client_secret
        current_app.config["NODE_KEY"] = self.node_key
        current_app.config["ENVIRONMENT"] = self.environment
        self.local_cleanup_only = local_cleanup_only
        self.root = root
        self.domain = None

        if not local_cleanup_only:
            if not (self.client_id and self.client_secret and self.node_key):
                click.echo("This node is not configured... exiting")
                click.get_current_context().exit(1)

    def run(self):
        if self.local_cleanup_only:
            tasks = [
                self.stop_services,
                self.remove_certs,
                self.remove_vhosts,
                self.cleanup_oidc_server,
                self.cleanup_local_database_no_key,
                self.cleanup_info_json,
            ]
        else:
            tasks = [
                self.stop_services,
                self.synchronize_state,
                self.remove_node,
                self.remove_certs,
                self.remove_vhosts,
                self.cleanup_oidc_server,
                self.cleanup_local_database,
                self.cleanup_info_json,
            ]

        def show(obj):
            if obj and obj.__doc__:
                return obj.__doc__.strip().split("\n")[0]
            return ""

        with click.progressbar(tasks, item_show_func=show, show_eta=False) as bar:
            for task in bar:
                task()

    def synchronize_state(self):
        "Synchronize Endpoint State"
        current_app.logger.info("synchronize_state: start")

        current_app.logger.debug("synchronize_state: download_keychain")
        try:
            items = current_app.appsync_client.query_all(
                id=ConfigurationObjectTag.KEY_CHAIN.value
            )
            if items:
                current_app.appsync_client.process_item(items[0])
        except globus_sdk.exc.AuthAPIError as e:
            if e.code == "UNAUTHORIZED":
                click.echo(
                    "\n".join(
                        textwrap.wrap(
                            "Error authenticating to Globus with the "
                            "endpoint's client credentials. This can "
                            "occur if the client_id "
                            "or secret has been entered incorrectly or if the "
                            "endpoint has been deleted."
                        )
                        + [""]
                        + textwrap.wrap(
                            "Verify the client_id and secret on "
                            "on {}".format(
                                globus_sdk.config.get_service_url(
                                    os.environ.get("GLOBUS_SDK_ENVIRONMENT", "default"),
                                    "auth",
                                )
                            )
                        )
                        + [""]
                        + textwrap.wrap(
                            "If the endpoint has been deleted, "
                            "you can still clean up the configuration on this "
                            "node using the --local-cleanup-only flag."
                        )
                    ),
                    err=True,
                )
                click.get_current_context().exit(1)
            else:
                raise

        current_app.logger.debug("synchronize_state: download_domain")
        items = current_app.appsync_client.query_all(
            tag=ConfigurationObjectTag.WILDCARD_DOMAIN
        )
        current_app.appsync_client.commit_items_to_session(db.session, items)
        current_app.logger.debug("synchronize_state: commit to database")
        current_app.logger.info("synchronize_state: end")

    def remove_node(self):
        "Remove Node"
        current_app.logger.info("remove_node: start")

        current_app.logger.debug("remove_node: query_node_id")
        node_conf = Configuration.query.filter(
            Configuration.name == "node_id"
        ).one_or_none()

        if node_conf:
            current_app.logger.debug("remove_node: node_id=%s", node_conf.value)

        current_app.logger.debug("remove_node: query_domain")
        domain = GlobusDnsDomain.query.one_or_none()

        if domain:
            if node_conf:
                current_app.logger.debug("remove_node: get_by_id")
                node = domain.get_node(id=node_conf.value)
                current_app.logger.debug("remove_node: node=%s", node)
            else:
                current_app.logger.debug("remove_node: get_by_ip_address")
                node = domain.get_node(ip_address=self.ip_address)
                current_app.logger.debug("remove_node: node=%s", node)
            if node:
                current_app.logger.debug("remove_node: remove_node id=%s", node.id)
                domain.remove_node(node.id)

            try:
                current_app.logger.debug("remove_node: synchronize")
                current_app.appsync_client.update_object(domain)
            except GraphQLError as e:
                current_app.logger.warning("remove_node: synchronize error: %s", str(e))
                current_app.logger.info("remove_node: rollback")
                domain.rollback(e.data[0])
                db.session.add(domain)
                db.session.commit()
                raise e
            db.session.add(domain)
            db.session.commit()
        current_app.logger.info("remove_node: end")

    def stop_services(self):
        "Stop services"
        current_app.logger.info("stop_services: start")
        self.disable_apache()
        gcs_manager = GCSManager()
        gcs_manager.cleanup()
        ServiceManager.disable("gcs_manager_assistant")
        ServiceManager.stop("gcs_manager_assistant")

        self.cleanup_gridftp_server()

        current_app.logger.info("stop_services: end")

    def cleanup_gridftp_server(self):
        current_app.logger.info("cleanup_gridftp_server: start")
        gridftp = GridFTPService(
            None,
            None,
            self.ip_address,
            self.domain,
            environment=self.environment or "default",
            root=self.root,
        )
        gridftp.cleanup()
        current_app.logger.debug("cleanup_gridftp_server: remove gridftp-key")
        gridftp_keyfile = os.path.expanduser("~gcsweb/gridftp-key")
        if os.path.exists(gridftp_keyfile):
            os.remove(gridftp_keyfile)
        if os.path.exists(gridftp_keyfile + ".old"):
            os.remove(gridftp_keyfile + ".old")
        dotglobus = os.path.expanduser("~gcsweb/.globus")
        if os.path.exists(dotglobus):
            shutil.rmtree(dotglobus, ignore_errors=True)
        current_app.logger.info("cleanup_gridftp_server: end")

    def disable_apache(self):
        current_app.logger.info("disable_apache: start")
        for module in WebServerManager.required_modules:
            WebServerManager.module_disable(module)
        for apacheconf in WebServerManager.required_confs:
            WebServerManager.conf_disable(apacheconf)
        current_app.logger.info("disable_apache: end")

    def remove_certs(self):
        current_app.logger.info("remove_certs: start")

        letsencryptdir = os.path.join(
            current_app.config["HOME"], "etc/letsencrypt/live"
        )

        if os.path.exists(letsencryptdir):
            for domain in os.listdir(letsencryptdir):
                livedir = os.path.join(letsencryptdir, domain)
                current_app.logger.debug("remove_certs: rmtree %s", livedir)
                shutil.rmtree(livedir, ignore_errors=True)

        custom_domain_dir = os.path.join(current_app.config["HOME"], "etc/domains")
        if os.path.exists(custom_domain_dir):
            for domain in os.listdir(custom_domain_dir):
                domaindir = os.path.join(custom_domain_dir, domain)
                shutil.rmtree(domaindir, ignore_errors=True)

        current_app.logger.info("remove_certs: end")

    def remove_vhosts(self):
        current_app.logger.info("remove_vhosts: start")
        vhostdir = os.path.join(current_app.config["HOME"], "etc/httpd/conf.d")
        if os.path.exists(vhostdir):
            for domain in os.listdir(vhostdir):
                vhost = os.path.join(vhostdir, domain)
                current_app.logger.debug("remove_vhosts: rm %s", vhost)
                os.remove(vhost)
        current_app.logger.info("remove_vhosts: end")

    def cleanup_local_database(self):
        "Clean up local database"
        current_app.logger.info("cleanup_local_database: start")
        for c in Collection.query.filter(Collection.is_guest_collection):
            for r in Role.query.filter(Role.resource == c.id):
                db.session.delete(r)
            current_app.logger.debug("cleanup_vhost for %s", c.id)
            c.cleanup_vhost()
            db.session.delete(c)
        for c in Collection.query.filter(Collection.is_mapped_collection):
            for r in Role.query.filter(Role.resource == c.id):
                db.session.delete(r)
            current_app.logger.debug("cleanup_vhost for %s", c.id)
            c.cleanup_vhost()
            db.session.delete(c)
        for c in UserCredential.query.all():
            db.session.delete(c)
        for c in StorageGateway.query.all():
            db.session.delete(c)
        for c in Role.query.all():
            db.session.delete(c)
        for c in LetsEncryptCertificate.query.all():
            c.write_config(delete=True)
            db.session.delete(c)
        for c in ConfigurationObject.query.all():
            db.session.delete(c)

        if current_app.config["NODE_KEY"]:
            current_app.keychain.remove_node_key(
                current_app.config["NODE_KEY"].thumbprint()
            )
            try:
                current_app.appsync_client.update_object(
                    current_app.keychain, ignore_response=True
                )
            except Exception as e:
                print("EXC: {}".format(e))

        for c in Configuration.query.all():
            if c.name != "secret":
                db.session.delete(c)
        db.session.commit()
        current_app.logger.info("cleanup_local_database: end")

    def cleanup_local_database_no_key(self):
        "Clean up local database"
        current_app.logger.info("cleanup_local_database_no_key: start")

        db.session.execute("PRAGMA foreign_keys = OFF")
        Role.query.delete()
        Collection.query.delete()
        UserCredential.query.delete()
        StorageGateway.query.delete()
        LetsEncryptCertificate.query.delete()
        ConfigurationObject.query.delete()
        Configuration.query.delete()
        db.session.commit()
        current_app.logger.info("cleanup_local_database: end")

    def cleanup_info_json(self):
        """
        Delete local info
        """
        current_app.logger.info("cleanup_info_json: start")
        info_json = os.path.join(self.root, "var/lib/globus-connect-server/info.json",)
        if os.path.exists(info_json):
            current_app.logger.debug("cleanup_info_json: remove %s", info_json)
            os.truncate(info_json, 0)
        current_app.logger.info("cleanup_info_json: end")

    def cleanup_oidc_server(self):
        """
        Cleanup the OIDC server
        """
        current_app.logger.info("cleanup_oidc_server: start")
        endpoint = Endpoint.query.one_or_none()
        if endpoint and endpoint.has_oidc_server:
            current_app.logger.debug("cleanup_oidc_server: deleting config")
            endpoint.oidc_server.write_config(delete=True)
            ServiceManager.disable("globus-oidc", path=True, socket=True)
            ServiceManager.stop("globus-oidc", path=True, socket=True)
            ServiceManager.stop("globus-oidc")
        current_app.logger.info("cleanup_oidc_server: end")
