import re
import socket
from typing import Optional
import urllib.request

import ifaddr
import ipaddress


def is_ec2() -> bool:
    """
    Returns True if the current machine appears to be an ec2 instance, False
    otherwise.
    """
    # This is based on the documentation from amazon here:
    # https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/identify_ec2_instances.html
    result = False
    hypervisor_uuidfile = "/sys/hypervisor/uuid"
    try:
        with open(hypervisor_uuidfile, "r") as f:
            result = f.read(3).lower() == "ec2"
    except Exception:
        pass

    if not result:
        # Check this for both big-endian and little-endian uuids
        # Only works when run as root, but is ok for node setup
        dmi_product_uuid_file = "/sys/devices/virtual/dmi/id/product_uuid"
        try:
            with open(dmi_product_uuid_file, "r") as f:
                uuid = f.read(36).lower()
                result = uuid[0:3] == "ec2" or (uuid[6:8] + uuid[4]) == "ec2"
        except Exception:
            pass

    return result


def public_ip() -> Optional[str]:
    """
    Determine a public IP address for this machine. This employs two different
    algorithms:

    *  If this is an EC2 instance, parse the metadata and extract the public
       IPv4 address
    *  Enumerate network interfaces and check for a single public IP address
    """
    value = None

    if is_ec2():
        url = "http://169.254.169.254/latest/meta-data/public-ipv4"
        try:
            socket.setdefaulttimeout(3.0)
            resource = urllib.request.urlopen(url)
            charset = resource.headers.get_content_charset()
            if not charset:
                charset = "ascii"
            value = resource.read().decode(charset)
        except IOError:
            pass

        if value is not None and "404 - Not Found" in value:
            value = None

    if value is None:
        # Enumerate network adapters and check for a public IPv4 address
        adapters = ifaddr.get_adapters()

        public_addresses = []

        for adapter in adapters:
            for ip in adapter.ips:
                if ip.is_IPv6:
                    continue

                address = ipaddress.ip_address(ip.ip)
                if address.is_global:
                    public_addresses.append(str(address))
        if len(public_addresses) == 1:
            value = public_addresses[0]

    return value


def is_private_ip(name: Optional[str]) -> bool:
    """
    Determine if a host name resolves to an ip address in a private
    block.
    """
    if name is None:
        return False
    elif re.match(
        r"^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}"
        + r"(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$",
        name,
    ):
        addr = name
    else:
        try:
            addr = socket.gethostbyname(name)
        except Exception:
            return True
    octets = [int(x) for x in addr.split(".")]
    return (
        octets[0] == 10
        or (octets[0] == 172 and octets[1] >= 16 and octets[1] <= 31)
        or (octets[0] == 192 and octets[1] == 168)
        or (octets[0] == 127)
    )
